# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: WhisperTextProtocol.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='WhisperTextProtocol.proto',
  package='textsecure',
  serialized_pb=_b('\n\x19WhisperTextProtocol.proto\x12\ntextsecure\"b\n\x0eWhisperMessage\x12\x12\n\nratchetKey\x18\x01 \x01(\x0c\x12\x0f\n\x07\x63ounter\x18\x02 \x01(\r\x12\x17\n\x0fpreviousCounter\x18\x03 \x01(\r\x12\x12\n\nciphertext\x18\x04 \x01(\x0c\"\x8f\x01\n\x14PreKeyWhisperMessage\x12\x16\n\x0eregistrationId\x18\x05 \x01(\r\x12\x10\n\x08preKeyId\x18\x01 \x01(\r\x12\x16\n\x0esignedPreKeyId\x18\x06 \x01(\r\x12\x0f\n\x07\x62\x61seKey\x18\x02 \x01(\x0c\x12\x13\n\x0bidentityKey\x18\x03 \x01(\x0c\x12\x0f\n\x07message\x18\x04 \x01(\x0c\"t\n\x12KeyExchangeMessage\x12\n\n\x02id\x18\x01 \x01(\r\x12\x0f\n\x07\x62\x61seKey\x18\x02 \x01(\x0c\x12\x12\n\nratchetKey\x18\x03 \x01(\x0c\x12\x13\n\x0bidentityKey\x18\x04 \x01(\x0c\x12\x18\n\x10\x62\x61seKeySignature\x18\x05 \x01(\x0c\"E\n\x10SenderKeyMessage\x12\n\n\x02id\x18\x01 \x01(\r\x12\x11\n\titeration\x18\x02 \x01(\r\x12\x12\n\nciphertext\x18\x03 \x01(\x0c\"c\n\x1cSenderKeyDistributionMessage\x12\n\n\x02id\x18\x01 \x01(\r\x12\x11\n\titeration\x18\x02 \x01(\r\x12\x10\n\x08\x63hainKey\x18\x03 \x01(\x0c\x12\x12\n\nsigningKey\x18\x04 \x01(\x0c\x42\x37\n&org.whispersystems.libaxolotl.protocolB\rWhisperProtos')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_WHISPERMESSAGE = _descriptor.Descriptor(
  name='WhisperMessage',
  full_name='textsecure.WhisperMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ratchetKey', full_name='textsecure.WhisperMessage.ratchetKey', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='counter', full_name='textsecure.WhisperMessage.counter', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='previousCounter', full_name='textsecure.WhisperMessage.previousCounter', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ciphertext', full_name='textsecure.WhisperMessage.ciphertext', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=41,
  serialized_end=139,
)


_PREKEYWHISPERMESSAGE = _descriptor.Descriptor(
  name='PreKeyWhisperMessage',
  full_name='textsecure.PreKeyWhisperMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='registrationId', full_name='textsecure.PreKeyWhisperMessage.registrationId', index=0,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='preKeyId', full_name='textsecure.PreKeyWhisperMessage.preKeyId', index=1,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='signedPreKeyId', full_name='textsecure.PreKeyWhisperMessage.signedPreKeyId', index=2,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='baseKey', full_name='textsecure.PreKeyWhisperMessage.baseKey', index=3,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='identityKey', full_name='textsecure.PreKeyWhisperMessage.identityKey', index=4,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='message', full_name='textsecure.PreKeyWhisperMessage.message', index=5,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=142,
  serialized_end=285,
)


_KEYEXCHANGEMESSAGE = _descriptor.Descriptor(
  name='KeyExchangeMessage',
  full_name='textsecure.KeyExchangeMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='textsecure.KeyExchangeMessage.id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='baseKey', full_name='textsecure.KeyExchangeMessage.baseKey', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ratchetKey', full_name='textsecure.KeyExchangeMessage.ratchetKey', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='identityKey', full_name='textsecure.KeyExchangeMessage.identityKey', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='baseKeySignature', full_name='textsecure.KeyExchangeMessage.baseKeySignature', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=287,
  serialized_end=403,
)


_SENDERKEYMESSAGE = _descriptor.Descriptor(
  name='SenderKeyMessage',
  full_name='textsecure.SenderKeyMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='textsecure.SenderKeyMessage.id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='iteration', full_name='textsecure.SenderKeyMessage.iteration', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ciphertext', full_name='textsecure.SenderKeyMessage.ciphertext', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=405,
  serialized_end=474,
)


_SENDERKEYDISTRIBUTIONMESSAGE = _descriptor.Descriptor(
  name='SenderKeyDistributionMessage',
  full_name='textsecure.SenderKeyDistributionMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='textsecure.SenderKeyDistributionMessage.id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='iteration', full_name='textsecure.SenderKeyDistributionMessage.iteration', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='chainKey', full_name='textsecure.SenderKeyDistributionMessage.chainKey', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='signingKey', full_name='textsecure.SenderKeyDistributionMessage.signingKey', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=476,
  serialized_end=575,
)

DESCRIPTOR.message_types_by_name['WhisperMessage'] = _WHISPERMESSAGE
DESCRIPTOR.message_types_by_name['PreKeyWhisperMessage'] = _PREKEYWHISPERMESSAGE
DESCRIPTOR.message_types_by_name['KeyExchangeMessage'] = _KEYEXCHANGEMESSAGE
DESCRIPTOR.message_types_by_name['SenderKeyMessage'] = _SENDERKEYMESSAGE
DESCRIPTOR.message_types_by_name['SenderKeyDistributionMessage'] = _SENDERKEYDISTRIBUTIONMESSAGE

WhisperMessage = _reflection.GeneratedProtocolMessageType('WhisperMessage', (_message.Message,), dict(
  DESCRIPTOR = _WHISPERMESSAGE,
  __module__ = 'WhisperTextProtocol_pb2'
  # @@protoc_insertion_point(class_scope:textsecure.WhisperMessage)
  ))
_sym_db.RegisterMessage(WhisperMessage)

PreKeyWhisperMessage = _reflection.GeneratedProtocolMessageType('PreKeyWhisperMessage', (_message.Message,), dict(
  DESCRIPTOR = _PREKEYWHISPERMESSAGE,
  __module__ = 'WhisperTextProtocol_pb2'
  # @@protoc_insertion_point(class_scope:textsecure.PreKeyWhisperMessage)
  ))
_sym_db.RegisterMessage(PreKeyWhisperMessage)

KeyExchangeMessage = _reflection.GeneratedProtocolMessageType('KeyExchangeMessage', (_message.Message,), dict(
  DESCRIPTOR = _KEYEXCHANGEMESSAGE,
  __module__ = 'WhisperTextProtocol_pb2'
  # @@protoc_insertion_point(class_scope:textsecure.KeyExchangeMessage)
  ))
_sym_db.RegisterMessage(KeyExchangeMessage)

SenderKeyMessage = _reflection.GeneratedProtocolMessageType('SenderKeyMessage', (_message.Message,), dict(
  DESCRIPTOR = _SENDERKEYMESSAGE,
  __module__ = 'WhisperTextProtocol_pb2'
  # @@protoc_insertion_point(class_scope:textsecure.SenderKeyMessage)
  ))
_sym_db.RegisterMessage(SenderKeyMessage)

SenderKeyDistributionMessage = _reflection.GeneratedProtocolMessageType('SenderKeyDistributionMessage', (_message.Message,), dict(
  DESCRIPTOR = _SENDERKEYDISTRIBUTIONMESSAGE,
  __module__ = 'WhisperTextProtocol_pb2'
  # @@protoc_insertion_point(class_scope:textsecure.SenderKeyDistributionMessage)
  ))
_sym_db.RegisterMessage(SenderKeyDistributionMessage)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('\n&org.whispersystems.libaxolotl.protocolB\rWhisperProtos'))
# @@protoc_insertion_point(module_scope)



import unittest
from axolotl.identitykey import IdentityKey
from axolotl.identitykeypair import IdentityKeyPair
from axolotl.ecc.curve import Curve
from axolotl.ecc.eckeypair import ECKeyPair
from axolotl.ratchet.bobaxolotlparamaters import BobAxolotlParameters
from axolotl.state.sessionstate import SessionState
from axolotl.ratchet.ratchetingsession import RatchetingSession
class RatchetingSessionTest(unittest.TestCase):
    def test_ratchetingSessionAsBob(self):
        bobPublic   = bytearray([0x05, 0x2c, 0xb4, 0x97,
                                    0x76, 0xb8, 0x77, 0x02,
                                    0x05, 0x74, 0x5a, 0x3a,
                                    0x6e, 0x24, 0xf5, 0x79,
                                    0xcd, 0xb4, 0xba, 0x7a,
                                    0x89, 0x04, 0x10, 0x05,
                                    0x92, 0x8e, 0xbb, 0xad,
                                    0xc9, 0xc0, 0x5a, 0xd4,
                                    0x58])

        bobPrivate  = bytearray([0xa1, 0xca, 0xb4, 0x8f,
                                    0x7c, 0x89, 0x3f, 0xaf,
                                    0xa9, 0x88, 0x0a, 0x28,
                                    0xc3, 0xb4, 0x99, 0x9d,
                                    0x28, 0xd6, 0x32, 0x95,
                                    0x62, 0xd2, 0x7a, 0x4e,
                                    0xa4, 0xe2, 0x2e, 0x9f,
                                    0xf1, 0xbd, 0xd6, 0x5a])

        bobIdentityPublic   = bytearray([0x05, 0xf1, 0xf4, 0x38,
                                    0x74, 0xf6, 0x96, 0x69,
                                    0x56, 0xc2, 0xdd, 0x47,
                                    0x3f, 0x8f, 0xa1, 0x5a,
                                    0xde, 0xb7, 0x1d, 0x1c,
                                    0xb9, 0x91, 0xb2, 0x34,
                                    0x16, 0x92, 0x32, 0x4c,
                                    0xef, 0xb1, 0xc5, 0xe6,
                                    0x26])

        bobIdentityPrivate    = bytearray([0x48, 0x75, 0xcc, 0x69,
                                    0xdd, 0xf8, 0xea, 0x07,
                                    0x19, 0xec, 0x94, 0x7d,
                                    0x61, 0x08, 0x11, 0x35,
                                    0x86, 0x8d, 0x5f, 0xd8,
                                    0x01, 0xf0, 0x2c, 0x02,
                                    0x25, 0xe5, 0x16, 0xdf,
                                    0x21, 0x56, 0x60, 0x5e])

        aliceBasePublic     = bytearray([0x05, 0x47, 0x2d, 0x1f,
                                    0xb1, 0xa9, 0x86, 0x2c,
                                    0x3a, 0xf6, 0xbe, 0xac,
                                    0xa8, 0x92, 0x02, 0x77,
                                    0xe2, 0xb2, 0x6f, 0x4a,
                                    0x79, 0x21, 0x3e, 0xc7,
                                    0xc9, 0x06, 0xae, 0xb3,
                                    0x5e, 0x03, 0xcf, 0x89,
                                    0x50])

        aliceEphemeralPublic  = bytearray([0x05, 0x6c, 0x3e, 0x0d,
                                    0x1f, 0x52, 0x02, 0x83,
                                    0xef, 0xcc, 0x55, 0xfc,
                                    0xa5, 0xe6, 0x70, 0x75,
                                    0xb9, 0x04, 0x00, 0x7f,
                                    0x18, 0x81, 0xd1, 0x51,
                                    0xaf, 0x76, 0xdf, 0x18,
                                    0xc5, 0x1d, 0x29, 0xd3,
                                    0x4b])

        aliceIdentityPublic   = bytearray([0x05, 0xb4, 0xa8, 0x45,
                                    0x56, 0x60, 0xad, 0xa6,
                                    0x5b, 0x40, 0x10, 0x07,
                                    0xf6, 0x15, 0xe6, 0x54,
                                    0x04, 0x17, 0x46, 0x43,
                                    0x2e, 0x33, 0x39, 0xc6,
                                    0x87, 0x51, 0x49, 0xbc,
                                    0xee, 0xfc, 0xb4, 0x2b,
                                    0x4a])

        senderChain           = bytearray([0xd2, 0x2f, 0xd5, 0x6d, 0x3f,
                                    0xec, 0x81, 0x9c, 0xf4, 0xc3,
                                    0xd5, 0x0c, 0x56, 0xed, 0xfb,
                                    0x1c, 0x28, 0x0a, 0x1b, 0x31,
                                    0x96, 0x45, 0x37, 0xf1, 0xd1,
                                    0x61, 0xe1, 0xc9, 0x31, 0x48,
                                    0xe3, 0x6b])

        bobIdentityKeyPublic   = IdentityKey(bobIdentityPublic, 0)
        bobIdentityKeyPrivate  = Curve.decodePrivatePoint(bobIdentityPrivate)
        bobIdentityKey         =  IdentityKeyPair(bobIdentityKeyPublic, bobIdentityKeyPrivate)
        bobEphemeralPublicKey  = Curve.decodePoint(bobPublic, 0)
        bobEphemeralPrivateKey = Curve.decodePrivatePoint(bobPrivate)
        bobEphemeralKey        = ECKeyPair(bobEphemeralPublicKey, bobEphemeralPrivateKey)
        bobBaseKey             = bobEphemeralKey

        aliceBasePublicKey       = Curve.decodePoint(aliceBasePublic, 0)
        aliceEphemeralPublicKey  = Curve.decodePoint(aliceEphemeralPublic, 0)
        aliceIdentityPublicKey   = IdentityKey(aliceIdentityPublic, 0)

        parameters = BobAxolotlParameters.newBuilder()\
        .setOurIdentityKey(bobIdentityKey)\
        .setOurSignedPreKey(bobBaseKey)\
        .setOurRatchetKey(bobEphemeralKey)\
        .setOurOneTimePreKey(None)\
        .setTheirIdentityKey(aliceIdentityPublicKey)\
        .setTheirBaseKey(aliceBasePublicKey)\
        .create()

        session = SessionState()

        RatchetingSession.initializeSessionAsBob(session, 2, parameters)
        self.assertEqual(session.getLocalIdentityKey(), bobIdentityKey.getPublicKey())
        self.assertEqual(session.getRemoteIdentityKey(), aliceIdentityPublicKey)
        self.assertEqual(session.getSenderChainKey().getKey(), senderChain)



class BobAxolotlParameters:

    def __init__(self, ourIdentityKey, ourSignedPreKey,
                 ourRatchetKey, ourOneTimePreKey,
                 theirIdentityKey, theirBaseKey):
        """
        :type ourIdentityKey: IdentityKeyPair
        :type ourSignedPreKey: ECKeyPair
        :type ourRatchetKey: ECKeyPair
        :type ourOneTimePreKey: ECKeyPair
        :type theirIdentityKey: IdentityKey
        :type theirBaseKey: ECPublicKey
        """

        self.ourIdentityKey = ourIdentityKey
        self.ourSignedPreKey = ourSignedPreKey
        self.ourRatchetKey = ourRatchetKey
        self.ourOneTimePreKey = ourOneTimePreKey
        self.theirIdentityKey = theirIdentityKey
        self.theirBaseKey = theirBaseKey

        if ourIdentityKey is None or ourSignedPreKey is None or ourRatchetKey is None\
            or theirIdentityKey is None or theirBaseKey is None:
            raise ValueError("Null value!")

    def getOurIdentityKey(self):
        return self.ourIdentityKey

    def getOurSignedPreKey(self):
        return self.ourSignedPreKey

    def getOurOneTimePreKey(self):
        return self.ourOneTimePreKey

    def getTheirIdentityKey(self):
        return self.theirIdentityKey

    def getTheirBaseKey(self):
        return self.theirBaseKey

    def getOurRatchetKey(self):
        return self.ourRatchetKey

    @staticmethod
    def newBuilder():
        return BobAxolotlParameters.Builder()

    class Builder:
        def __init__(self):
            self.ourIdentityKey = None
            self.ourSignedPreKey = None
            self.ourOneTimePreKey = None
            self.ourRatchetKey = None
            self.theirIdentityKey = None
            self.theirBaseKey = None

        def setOurIdentityKey(self, ourIdentityKey):
            self.ourIdentityKey = ourIdentityKey
            return self

        def setOurSignedPreKey(self, ourSignedPreKey):
            self.ourSignedPreKey = ourSignedPreKey
            return self

        def setOurOneTimePreKey(self, ourOneTimePreKey):
            self.ourOneTimePreKey = ourOneTimePreKey
            return self

        def setOurRatchetKey(self, ourRatchetKey):
            self.ourRatchetKey = ourRatchetKey
            return self

        def setTheirIdentityKey(self, theirIdentityKey):
            self.theirIdentityKey =theirIdentityKey
            return self

        def setTheirBaseKey(self, theirBaseKey):
            self.theirBaseKey = theirBaseKey
            return self

        def create(self):
            return BobAxolotlParameters(self.ourIdentityKey, self.ourSignedPreKey, self.ourRatchetKey,
                                        self.ourOneTimePreKey, self.theirIdentityKey, self.theirBaseKey)






import abc
class SignedPreKeyStore(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def loadSignedPreKey(self, signedPreKeyId):pass

    @abc.abstractmethod
    def loadSignedPreKeys(self):pass

    @abc.abstractmethod
    def storeSignedPreKey(self, signedPreKeyId, signedPreKeyRecord): pass

    @abc.abstractmethod
    def containsSignedPreKey(self, signedPreKeyId): pass

    @abc.abstractmethod
    def removeSignedPreKey(self, signedPreKeyId): pass


class ByteUtil:

    @staticmethod
    def combine(*args):
        baos = bytearray()
        for a in args:
            if type(a) in (list, bytearray, str, bytes):
                baos.extend(a)
            else:
                baos.append(a)

        return baos

    # @staticmethod
    # def xsplit(inp, firstLength, secondLength, thirdLength = None):
    #     parts = []
    #     parts.append(inp[:firstLength])
    #     parts.append(inp[len(parts[0]): secondLength + 1])
    #     if thirdLength:
    #         parts.append(inp[len(parts[1]): thirdLength + 1])
    #     return parts

    @staticmethod
    def split(inp, firstLength, secondLength, thirdLength = None):
        parts = []
        parts.append(inp[:firstLength])
        parts.append(inp[firstLength:firstLength + secondLength])
        if thirdLength is not None:
            parts.append(inp[firstLength + secondLength: firstLength + secondLength + thirdLength])

        return parts


    @staticmethod
    def trim(inp, length):
        return inp[:length]

    @staticmethod
    def intsToByteHighAndLow(highValue, lowValue):
        highValue = ord(highValue) if type(highValue) is str else highValue
        lowValue = ord(lowValue) if type(lowValue) is str else lowValue
        return ((highValue << 4 | lowValue) & 0xFF) % 256

    @staticmethod
    def highBitsToInt(value):
        bit = ord(value) if type(value) is str else value
        return (bit & 0xFF) >> 4

    @staticmethod
    def lowBitsToInt(value):
        return value & 0xF

    @staticmethod
    def intToByteArray(_bytes, offset, value):
        _bytes[offset + 3] = value % 256
        _bytes[offset + 2] = (value >> 8) % 256
        _bytes[offset + 1] = (value >> 16) % 256
        _bytes[offset] = (value >> 24) % 256

        return 4



from ..ecc.curve import Curve
from ..identitykey import IdentityKey
from ..identitykeypair import IdentityKeyPair
from ..state.prekeyrecord import PreKeyRecord
from ..state.signedprekeyrecord import SignedPreKeyRecord
from .medium import Medium
import os
import struct
import time
import binascii
import math

class KeyHelper:

    def __init__(self):
        pass

    """
    Generate an identity key pair.  Clients should only do this once,
    at install time.
    @return the generated IdentityKeyPair.
    """
    @staticmethod
    def generateIdentityKeyPair():
        keyPair   = Curve.generateKeyPair()
        publicKey = IdentityKey(keyPair.getPublicKey())
        serialized = '0a21056e8936e8367f768a7bba008ade7cf58407bdc7a6aae293e2cb7c06668dcd7d5e12205011524f0c15467100dd6' \
                     '03e0d6020f4d293edfbcd82129b14a88791ac81365c'
        serialized = binascii.unhexlify(serialized.encode())
        identityKeyPair = IdentityKeyPair(publicKey, keyPair.getPrivateKey())
        return identityKeyPair
        # return IdentityKeyPair(serialized=serialized)

    """
    Generate a registration ID.  Clients should only do this once,
    at install time.
    """
    @staticmethod
    def generateRegistrationId():
        regId =  KeyHelper.getRandomSequence()
        return regId

    @staticmethod
    def getRandomSequence(max = 4294967296):
        size = int(math.log(max)/ math.log(2)) / 8
        rand = os.urandom(int(size))
        randh = binascii.hexlify(rand)
        return int(randh, 16)

    """
    Generate a list of PreKeys.  Clients should do this at install time, and
    subsequently any time the list of PreKeys stored on the server runs low.

    PreKey IDs are shorts, so they will eventually be repeated.  Clients should
    store PreKeys in a circular buffer, so that they are repeated as infrequently
    as possible.

    @param start The starting PreKey ID, inclusive.
    @param count The number of PreKeys to generate.
    @return the list of generated PreKeyRecords.
   """
    @staticmethod
    def generatePreKeys(start, count):
        results = []
        start -= 1
        for i in range(0, count):
            preKeyId = ((start + i) % (Medium.MAX_VALUE-1)) + 1
            results.append(PreKeyRecord(preKeyId, Curve.generateKeyPair()))

        return results

    @staticmethod
    def generateSignedPreKey(identityKeyPair, signedPreKeyId):
        keyPair = Curve.generateKeyPair()
        signature = Curve.calculateSignature(identityKeyPair.getPrivateKey(), keyPair.getPublicKey().serialize())

        spk = SignedPreKeyRecord(signedPreKeyId, int(round(time.time() * 1000)), keyPair, signature)

        return spk

    @staticmethod
    def generateSenderSigningKey():
        return Curve.generateKeyPair()

    @staticmethod
    def generateSenderKey():
        return os.urandom(32)

    @staticmethod
    def generateSenderKeyId():
        return KeyHelper.getRandomSequence(2147483647)


import abc
class SessionStore(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def loadSession(self, recepientId, deviceId):pass

    @abc.abstractmethod
    def getSubDeviceSessions(self, recepientId): pass

    @abc.abstractmethod
    def storeSession(self, recepientId, deviceId, sessionRecord):pass

    @abc.abstractmethod
    def containsSession(self, recepientId, deviceId): pass

    @abc.abstractmethod
    def deleteSession(self, recepientId, deviceId): pass

    @abc.abstractmethod
    def deleteAllSessions(self, recepientId): pass


import abc
class PreKeyStore(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def loadPreKey(self, preKeyId):pass

    @abc.abstractmethod
    def storePreKey(self, preKeyId, preKeyRecord):pass

    @abc.abstractmethod
    def containsPreKey(self, preKeyId):pass

    @abc.abstractmethod
    def removePreKey(self, preKeyId):pass


from .sendermessagekey import SenderMessageKey
import hashlib
import hmac
class SenderChainKey:
    MESSAGE_KEY_SEED = bytearray([0x01])
    CHAIN_KEY_SEED = bytearray([0x02])

    def __init__(self, iteration, chainKey):
        """
        :type iteration: int
        :type chainKey: bytearray
        """
        self.iteration = iteration
        self.chainKey = chainKey

    def getIteration(self):
        return self.iteration

    def getSenderMessageKey(self):
        return SenderMessageKey(self.iteration, self.getDerivative(self.__class__.MESSAGE_KEY_SEED, self.chainKey))

    def getNext(self):
        return SenderChainKey(self.iteration + 1, self.getDerivative(self.__class__.CHAIN_KEY_SEED, self.chainKey))

    def getSeed(self):
        return self.chainKey

    def getDerivative(self, seed, key):
        mac = hmac.new(bytes(key), bytes(seed), digestmod = hashlib.sha256)
        return mac.digest()



import unittest
from axolotl.ecc.curve import Curve
from axolotl.util.keyhelper import KeyHelper
class Curve25519Test(unittest.TestCase):
    def test_agreement(self):
        alicePublic  =  bytearray([0x05, 0x1b,  0xb7,  0x59,  0x66,
                        0xf2,  0xe9,  0x3a,  0x36,  0x91,
                        0xdf,  0xff,  0x94,  0x2b,  0xb2,
                        0xa4,  0x66,  0xa1,  0xc0,  0x8b,
                        0x8d,  0x78,  0xca,  0x3f,  0x4d,
                        0x6d,  0xf8,  0xb8,  0xbf,  0xa2,
                        0xe4,  0xee,  0x28])
        alicePrivate = bytearray([0xc8, 0x06, 0x43, 0x9d, 0xc9,
                           0xd2, 0xc4, 0x76, 0xff, 0xed,
                           0x8f, 0x25, 0x80, 0xc0, 0x88,
                           0x8d, 0x58, 0xab, 0x40, 0x6b,
                           0xf7, 0xae, 0x36, 0x98, 0x87,
                           0x90, 0x21, 0xb9, 0x6b, 0xb4,
                           0xbf, 0x59])

        bobPublic    = bytearray([0x05, 0x65, 0x36, 0x14, 0x99,
                           0x3d, 0x2b, 0x15, 0xee, 0x9e,
                           0x5f, 0xd3, 0xd8, 0x6c, 0xe7,
                           0x19, 0xef, 0x4e, 0xc1, 0xda,
                           0xae, 0x18, 0x86, 0xa8, 0x7b,
                           0x3f, 0x5f, 0xa9, 0x56, 0x5a,
                           0x27, 0xa2, 0x2f])

        bobPrivate   = bytearray([0xb0, 0x3b, 0x34, 0xc3, 0x3a,
                           0x1c, 0x44, 0xf2, 0x25, 0xb6,
                           0x62, 0xd2, 0xbf, 0x48, 0x59,
                           0xb8, 0x13, 0x54, 0x11, 0xfa,
                           0x7b, 0x03, 0x86, 0xd4, 0x5f,
                           0xb7, 0x5d, 0xc5, 0xb9, 0x1b,
                           0x44, 0x66])

        shared       = bytearray([0x32, 0x5f, 0x23, 0x93, 0x28,
                           0x94, 0x1c, 0xed, 0x6e, 0x67,
                           0x3b, 0x86, 0xba, 0x41, 0x01,
                           0x74, 0x48, 0xe9, 0x9b, 0x64,
                           0x9a, 0x9c, 0x38, 0x06, 0xc1,
                           0xdd, 0x7c, 0xa4, 0xc4, 0x77,
                           0xe6, 0x29])

        alicePublicKey = Curve.decodePoint(alicePublic, 0)
        alicePrivateKey = Curve.decodePrivatePoint(alicePrivate)

        bobPublicKey = Curve.decodePoint(bobPublic, 0)
        bobPrivateKey = Curve.decodePrivatePoint(bobPrivate)

        sharedOne = Curve.calculateAgreement(alicePublicKey, bobPrivateKey)
        sharedTwo = Curve.calculateAgreement(bobPublicKey, alicePrivateKey)

        self.assertEqual(sharedOne, shared)
        self.assertEqual(sharedTwo, shared)

    def test_randomAgreements(self):

        for i in range(0, 50):
            alice   = Curve.generateKeyPair()
            bob     = Curve.generateKeyPair()
            sharedAlice = Curve.calculateAgreement(bob.getPublicKey(), alice.getPrivateKey())
            sharedBob   = Curve.calculateAgreement(alice.getPublicKey(), bob.getPrivateKey())
            self.assertEqual(sharedAlice, sharedBob)

    def test_gensig(self):
        identityKeyPair = KeyHelper.generateIdentityKeyPair()
        KeyHelper.generateSignedPreKey(identityKeyPair, 0)


    def test_signature(self):
        aliceIdentityPrivate = bytearray([0xc0, 0x97, 0x24, 0x84, 0x12,
                                   0xe5, 0x8b, 0xf0, 0x5d, 0xf4,
                                   0x87, 0x96, 0x82, 0x05, 0x13,
                                   0x27, 0x94, 0x17, 0x8e, 0x36,
                                   0x76, 0x37, 0xf5, 0x81, 0x8f,
                                   0x81, 0xe0, 0xe6, 0xce, 0x73,
                                   0xe8, 0x65])

        aliceIdentityPublic  = bytearray([0x05, 0xab, 0x7e, 0x71, 0x7d,
                                   0x4a, 0x16, 0x3b, 0x7d, 0x9a,
                                   0x1d, 0x80, 0x71, 0xdf, 0xe9,
                                   0xdc, 0xf8, 0xcd, 0xcd, 0x1c,
                                   0xea, 0x33, 0x39, 0xb6, 0x35,
                                   0x6b, 0xe8, 0x4d, 0x88, 0x7e,
                                   0x32, 0x2c, 0x64])

        aliceEphemeralPublic = bytearray([0x05, 0xed, 0xce, 0x9d, 0x9c,
                                   0x41, 0x5c, 0xa7, 0x8c, 0xb7,
                                   0x25, 0x2e, 0x72, 0xc2, 0xc4,
                                   0xa5, 0x54, 0xd3, 0xeb, 0x29,
                                   0x48, 0x5a, 0x0e, 0x1d, 0x50,
                                   0x31, 0x18, 0xd1, 0xa8, 0x2d,
                                   0x99, 0xfb, 0x4a])

        aliceSignature       = bytearray([0x5d, 0xe8, 0x8c, 0xa9, 0xa8,
                                   0x9b, 0x4a, 0x11, 0x5d, 0xa7,
                                   0x91, 0x09, 0xc6, 0x7c, 0x9c,
                                   0x74, 0x64, 0xa3, 0xe4, 0x18,
                                   0x02, 0x74, 0xf1, 0xcb, 0x8c,
                                   0x63, 0xc2, 0x98, 0x4e, 0x28,
                                   0x6d, 0xfb, 0xed, 0xe8, 0x2d,
                                   0xeb, 0x9d, 0xcd, 0x9f, 0xae,
                                   0x0b, 0xfb, 0xb8, 0x21, 0x56,
                                   0x9b, 0x3d, 0x90, 0x01, 0xbd,
                                   0x81, 0x30, 0xcd, 0x11, 0xd4,
                                   0x86, 0xce, 0xf0, 0x47, 0xbd,
                                   0x60, 0xb8, 0x6e, 0x88])

        alicePrivateKey = Curve.decodePrivatePoint(aliceIdentityPrivate)
        alicePublicKey  = Curve.decodePoint(aliceIdentityPublic, 0)
        aliceEphemeral  = Curve.decodePoint(aliceEphemeralPublic, 0)

        res = Curve.verifySignature(alicePublicKey, aliceEphemeral.serialize(), bytes(aliceSignature))
        self.assertTrue(res)


import abc
import hmac
import hashlib
import math
class HKDF(object):
    __metaclass__ = abc.ABCMeta
    HASH_OUTPUT_SIZE  = 32

    @staticmethod
    def createFor(messageVersion):
        from . import hkdfv2, hkdfv3
        if messageVersion == 2:
            return hkdfv2.HKDFv2()
        elif messageVersion == 3:
            return hkdfv3.HKDFv3()
        else:
            raise  AssertionError("Unknown version: %s " % messageVersion)

    def deriveSecrets(self, inputKeyMaterial, info, outputLength, salt = None):
        salt = salt or bytearray(self.__class__.HASH_OUTPUT_SIZE)
        prk = self.extract(salt, inputKeyMaterial)
        return self.expand(prk, info, outputLength)

    def extract(self, salt, inputKeyMaterial):
        mac = hmac.new(bytes(salt), digestmod=hashlib.sha256)
        mac.update(bytes(inputKeyMaterial))
        return mac.digest()

    def expand(self, prk, info, outputSize):

        iterations = int(math.ceil(float(outputSize)/ float(self.__class__.HASH_OUTPUT_SIZE)))
        mixin = bytearray()
        results = bytearray()
        remainingBytes = outputSize

        for i in range(self.getIterationStartOffset(), iterations + self.getIterationStartOffset()):
            mac = hmac.new(prk, digestmod=hashlib.sha256)
            mac.update(bytes(mixin))
            if info is not None:
                mac.update(bytes(info))
            updateChr = chr(i % 256)
            mac.update(updateChr.encode())

            stepResult = mac.digest()
            stepSize = min(remainingBytes, len(stepResult))
            results.extend(stepResult[:stepSize])
            mixin = stepResult
            remainingBytes -= stepSize

        return bytes(results)

    @abc.abstractmethod
    def getIterationStartOffset(self):
        return 0



from ..util.byteutil import ByteUtil
class DerivedMessageSecrets:
    SIZE              = 80
    CIPHER_KEY_LENGTH = 32
    MAC_KEY_LENGTH    = 32
    IV_LENGTH         = 16

    def __init__(self, okm):
        keys = ByteUtil.split(okm, self.__class__.CIPHER_KEY_LENGTH, self.__class__.MAC_KEY_LENGTH,
                              self.__class__.IV_LENGTH)
        self.cipherKey = keys[0] #AES
        self.macKey = keys[1] #sha256
        self.iv = keys[2]


    def getCipherKey(self):
        return self.cipherKey

    def getMacKey(self):
        return self.macKey

    def getIv(self):
        return self.iv



from .ciphertextmessage import CiphertextMessage
from axolotl.util.byteutil import ByteUtil
from axolotl.legacymessageexception import LegacyMessageException
from axolotl.invalidmessageexception import InvalidMessageException
from axolotl.invalidkeyexception import InvalidKeyException
from axolotl.ecc.curve import Curve
from . import whisperprotos

class SenderKeyMessage(CiphertextMessage):
    SIGNATURE_LENGTH = 64

    def __init__(self, keyId = None, iteration = None, ciphertext = None, signatureKey = None, serialized = None):
        assert bool(keyId is not None and iteration is not None and ciphertext is not None and signatureKey is not None)\
               ^ bool(serialized),\
            "Either pass arguments or serialized data"

        if serialized:
            try:
                messageParts = ByteUtil.split(serialized, 1, len(serialized)- 1 - self.__class__.SIGNATURE_LENGTH,
                                              self.__class__.SIGNATURE_LENGTH)

                version      = messageParts[0][0]
                message      = messageParts[1]
                signature    = messageParts[2]

                if ByteUtil.highBitsToInt(version) < 3:
                    raise LegacyMessageException("Legacy message: %s" % ByteUtil.highBitsToInt(version))

                if ByteUtil.highBitsToInt(version) > self.__class__.CURRENT_VERSION:
                    raise InvalidMessageException("Unknown version: %s" % ByteUtil.highBitsToInt(version))

                senderKeyMessage = whisperprotos.SenderKeyMessage()
                senderKeyMessage.ParseFromString(message)

                if senderKeyMessage.id is None or senderKeyMessage.iteration is None\
                    or senderKeyMessage.ciphertext is None:
                    raise InvalidMessageException("Incomplete message")

                self.serialized = serialized
                self.messageVersion = ByteUtil.highBitsToInt(version)

                self.keyId          = senderKeyMessage.id
                self.iteration      = senderKeyMessage.iteration
                self.ciphertext     = senderKeyMessage.ciphertext
            except Exception as e:
                raise InvalidMessageException(e)
        else:
            version = [ByteUtil.intsToByteHighAndLow(self.__class__.CURRENT_VERSION, self.__class__.CURRENT_VERSION)]

            message = whisperprotos.SenderKeyMessage()
            message.id = keyId
            message.iteration = iteration
            message.ciphertext = ciphertext
            message = message.SerializeToString()

            signature = self.getSignature(signatureKey, bytes(ByteUtil.combine(version, message)))

            self.serialized       = ByteUtil.combine(version, message, signature)
            self.messageVersion   = self.__class__.CURRENT_VERSION
            self.keyId            = keyId
            self.iteration        = iteration
            self.ciphertext       = ciphertext

    def getKeyId(self):
        return self.keyId

    def getIteration(self):
        return self.iteration

    def getCipherText(self):
        return self.ciphertext

    def verifySignature(self, signatureKey):
        """
        :type signatureKey: ECPublicKey
        """
        try:
            parts = ByteUtil.split(self.serialized, len(self.serialized)- self.__class__.SIGNATURE_LENGTH, self.__class__.SIGNATURE_LENGTH)

            if not Curve.verifySignature(signatureKey, parts[0], parts[1]):
                raise InvalidMessageException("Invalid signature!")
        except InvalidKeyException as e:
            raise InvalidMessageException(e)

    def getSignature(self, signatureKey, serialized):
        """
        :type signatureKey: ECPirvateKey
        :type serialized: bytearray
        """
        try:
            return Curve.calculateSignature(signatureKey, serialized)
        except InvalidKeyException as e:
            raise AssertionError(e)

    def serialize(self):
        return self.serialized

    def getType(self):
        return CiphertextMessage.SENDERKEY_TYPE



from axolotl.groups.state.senderkeystore import SenderKeyStore
from axolotl.groups.state.senderkeyrecord import SenderKeyRecord
class InMemorySenderKeyStore(SenderKeyStore):
    def __init__(self):
        self.store = {}

    def storeSenderKey(self, senderKeyId, senderKeyRecord):
        self.store[senderKeyId] = senderKeyRecord

    def loadSenderKey(self, senderKeyId):
        if senderKeyId in self.store:
            return SenderKeyRecord(serialized=self.store[senderKeyId].serialize())

        return SenderKeyRecord()



class AliceAxolotlParameters:
    def __init__(self,ourIdentityKey, ourBaseKey,
                 theirIdentityKey, theirSignedPreKey,
                  theirRatchetKey, theirOneTimePreKey):
        """
        :type ourBaseKey: ECKeyPair
        :type theirSignedPreKey: ECKeyPair
        :type ourIdentityKey: IdentityKeyPair
        :type theirOneTimePreKey: ECPublicKey
        :type theirRatchetKey: ECKeyPair
        :type theirIdentityKey: IdentityKey
        """

        self.ourBaseKey = ourBaseKey
        self.ourIdentityKey = ourIdentityKey
        self.theirSignedPreKey = theirSignedPreKey
        self.theirRatchetKey = theirRatchetKey
        self.theirIdentityKey = theirIdentityKey
        self.theirOneTimePreKey = theirOneTimePreKey

        if ourBaseKey is None or ourIdentityKey is None or theirSignedPreKey is None\
            or theirRatchetKey is None or theirIdentityKey is None or theirSignedPreKey is None:
            raise ValueError("Null value!")


    def getOurIdentityKey(self):
        return self.ourIdentityKey

    def getOurBaseKey(self):
        return self.ourBaseKey

    def getTheirIdentityKey(self):
        return self.theirIdentityKey

    def getTheirSignedPreKey(self):
        return self.theirSignedPreKey


    def getTheirOneTimePreKey(self):
        return self.theirOneTimePreKey

    def getTheirRatchetKey(self):
        return self.theirRatchetKey

    @staticmethod
    def newBuilder():
        return AliceAxolotlParameters.Builder()

    class Builder:
        def __init__(self):
            self.ourIdentityKey = None
            self.ourBaseKey = None
            self.theirIdentityKey = None
            self.theirSignedPreKey = None
            self.theirRatchetKey = None
            self.theirOneTimePreKey = None

        def setOurIdentityKey(self, ourIdentityKey):
            self.ourIdentityKey = ourIdentityKey
            return self

        def setOurBaseKey(self, ourBaseKey):
            self.ourBaseKey = ourBaseKey
            return self


        def setTheirRatchetKey(self, theirRatchetKey):
            self.theirRatchetKey = theirRatchetKey
            return self

        def setTheirIdentityKey(self, theirIdentityKey):
            self.theirIdentityKey = theirIdentityKey
            return self

        def setTheirSignedPreKey(self, theirSignedPreKey):
            self.theirSignedPreKey = theirSignedPreKey
            return self


        def setTheirOneTimePreKey(self, theirOneTimePreKey):
            self.theirOneTimePreKey = theirOneTimePreKey
            return self

        def create(self):
            return AliceAxolotlParameters(self.ourIdentityKey, self.ourBaseKey, self.theirIdentityKey,
                                          self.theirSignedPreKey, self.theirRatchetKey, self.theirOneTimePreKey)


from axolotl.state.signedprekeystore import SignedPreKeyStore
from axolotl.state.signedprekeyrecord import SignedPreKeyRecord
from axolotl.invalidkeyidexception import InvalidKeyIdException
class InMemorySignedPreKeyStore(SignedPreKeyStore):
    def __init__(self):
        self.store = {}

    def loadSignedPreKey(self, signedPreKeyId):
        if not signedPreKeyId in self.store:
            raise InvalidKeyIdException("No such signedprekeyrecord! %s " % signedPreKeyId)

        return SignedPreKeyRecord(serialized=self.store[signedPreKeyId])

    def loadSignedPreKeys(self):
        results = []
        for serialized in self.store.values():
            results.append(SignedPreKeyRecord(serialized=serialized))

        return results

    def storeSignedPreKey(self, signedPreKeyId, signedPreKeyRecord):
        self.store[signedPreKeyId] = signedPreKeyRecord.serialize()

    def containsSignedPreKey(self, signedPreKeyId):
        return signedPreKeyId in self.store

    def removeSignedPreKey(self, signedPreKeyId):
        del self.store[signedPreKeyId]


import hmac
import hashlib
from ..kdf.derivedmessagesecrets import DerivedMessageSecrets
from ..kdf.messagekeys import MessageKeys
class ChainKey:
    MESSAGE_KEY_SEED    = bytearray([0x01])
    CHAIN_KEY_SEED      = bytearray([0x02])

    def __init__(self, kdf, key, index):
        self.kdf = kdf
        self.key = key
        self.index = index

    def getKey(self):
        return self.key

    def getIndex(self):
        return self.index

    def getNextChainKey(self):
        nextKey = self.getBaseMaterial(self.__class__.CHAIN_KEY_SEED)
        return ChainKey(self.kdf, nextKey, self.index + 1)

    def getMessageKeys(self):
        inputKeyMaterial = self.getBaseMaterial(self.__class__.MESSAGE_KEY_SEED)
        keyMaterialBytes = self.kdf.deriveSecrets(inputKeyMaterial, bytearray("WhisperMessageKeys".encode()), DerivedMessageSecrets.SIZE)
        keyMaterial = DerivedMessageSecrets(keyMaterialBytes)
        return MessageKeys(keyMaterial.getCipherKey(), keyMaterial.getMacKey(), keyMaterial.getIv(), self.index)

    def getBaseMaterial(self, seedBytes):
        mac = hmac.new(bytes(self.key), digestmod=hashlib.sha256)
        mac.update(bytes(seedBytes))
        return mac.digest()




from .state.storageprotos import IdentityKeyPairStructure
from .identitykey import IdentityKey
from .ecc.curve import Curve
class IdentityKeyPair:
    def __init__(self, identityKeyPublicKey = None, ecPrivateKey = None, serialized = None):
        if serialized:
            structure = IdentityKeyPairStructure()
            structure.ParseFromString(serialized)
            self.publicKey = IdentityKey(bytearray(structure.publicKey), offset = 0)
            self.privateKey = Curve.decodePrivatePoint(bytearray(structure.privateKey))
        else:
            self.publicKey = identityKeyPublicKey
            self.privateKey = ecPrivateKey

    def getPublicKey(self):
        return self.publicKey

    def getPrivateKey(self):
        return self.privateKey

    def serialize(self):
        structure = IdentityKeyPairStructure()
        structure.publicKey = self.publicKey.serialize()
        structure.privateKey = self.privateKey.serialize()
        return structure.SerializeToString()


import unittest
from axolotl.ratchet.chainkey import ChainKey
from axolotl.kdf.hkdf import HKDF
class ChainKeyTest(unittest.TestCase):
    def test_chainKeyDerivationV2(self):
        seed = bytearray([0x8a, 0xb7, 0x2d, 0x6f, 0x4c,
                           0xc5, 0xac, 0x0d, 0x38, 0x7e,
                           0xaf, 0x46, 0x33, 0x78, 0xdd,
                           0xb2, 0x8e, 0xdd, 0x07, 0x38,
                           0x5b, 0x1c, 0xb0, 0x12, 0x50,
                           0xc7, 0x15, 0x98, 0x2e, 0x7a,
                           0xd4, 0x8f])

        messageKey   = bytearray([0x02, 0xa9, 0xaa, 0x6c, 0x7d,
                           0xbd, 0x64, 0xf9, 0xd3, 0xaa,
                           0x92, 0xf9, 0x2a, 0x27, 0x7b,
                           0xf5, 0x46, 0x09, 0xda, 0xdf,
                           0x0b, 0x00, 0x82, 0x8a, 0xcf,
                           0xc6, 0x1e, 0x3c, 0x72, 0x4b,
                           0x84, 0xa7])

        macKey       = bytearray([0xbf, 0xbe, 0x5e, 0xfb, 0x60,
                           0x30, 0x30, 0x52, 0x67, 0x42,
                           0xe3, 0xee, 0x89, 0xc7, 0x02,
                           0x4e, 0x88, 0x4e, 0x44, 0x0f,
                           0x1f, 0xf3, 0x76, 0xbb, 0x23,
                           0x17, 0xb2, 0xd6, 0x4d, 0xeb,
                           0x7c, 0x83])

        nextChainKey = bytearray([0x28, 0xe8, 0xf8, 0xfe, 0xe5,
                           0x4b, 0x80, 0x1e, 0xef, 0x7c,
                           0x5c, 0xfb, 0x2f, 0x17, 0xf3,
                           0x2c, 0x7b, 0x33, 0x44, 0x85,
                           0xbb, 0xb7, 0x0f, 0xac, 0x6e,
                           0xc1, 0x03, 0x42, 0xa2, 0x46,
                           0xd1, 0x5d])

        chainKey =  ChainKey(HKDF.createFor(2), seed, 0)
        self.assertEqual(chainKey.getKey(), seed)
        self.assertEqual(chainKey.getMessageKeys().getCipherKey(), messageKey)
        self.assertEqual(chainKey.getMessageKeys().getMacKey(), macKey)
        self.assertEqual(chainKey.getNextChainKey().getKey(), nextChainKey)
        self.assertEqual(chainKey.getIndex(), 0)
        self.assertEqual(chainKey.getMessageKeys().getCounter(), 0)
        self.assertEqual(chainKey.getNextChainKey().getIndex(), 1)
        self.assertEqual(chainKey.getNextChainKey().getMessageKeys().getCounter(), 1)




class InvalidMessageException(Exception):
    def __init__(self, message, exceptions = None):
        if exceptions:
            message = message + "%s" % exceptions[0]
        super(InvalidMessageException, self).__init__(message)


from ..ecc.curve import Curve
from ..kdf.derivedrootsecrets import DerivedRootSecrets
from .chainkey import ChainKey
class RootKey:
    def __init__(self, kdf, key):
        self.kdf = kdf
        self.key = key

    def getKeyBytes(self):
        return self.key

    def createChain(self, ECPublicKey_theirRatchetKey, ECKeyPair_ourRatchetKey):
        sharedSecret = Curve.calculateAgreement(ECPublicKey_theirRatchetKey, ECKeyPair_ourRatchetKey.getPrivateKey())
        derivedSecretBytes = self.kdf.deriveSecrets(sharedSecret, bytearray("WhisperRatchet".encode()), DerivedRootSecrets.SIZE, salt = self.key)
        derivedSecrets = DerivedRootSecrets(derivedSecretBytes)
        newRootKey = RootKey(self.kdf, derivedSecrets.getRootKey())
        newChainKey = ChainKey(self.kdf, derivedSecrets.getChainKey(), 0)
        return (newRootKey, newChainKey)



# coding=utf-8
import unittest
import time
from axolotl.sessionbuilder import SessionBuilder
from axolotl.sessioncipher import SessionCipher
from axolotl.ecc.curve import Curve
from axolotl.protocol.ciphertextmessage import CiphertextMessage
from axolotl.protocol.whispermessage import WhisperMessage
from axolotl.protocol.prekeywhispermessage import PreKeyWhisperMessage
from axolotl.state.prekeybundle import PreKeyBundle
from axolotl.tests.inmemoryaxolotlstore import InMemoryAxolotlStore
from axolotl.state.prekeyrecord import PreKeyRecord
from axolotl.state.signedprekeyrecord import SignedPreKeyRecord
from axolotl.tests.inmemoryidentitykeystore import InMemoryIdentityKeyStore
from axolotl.protocol.keyexchangemessage import KeyExchangeMessage
from axolotl.untrustedidentityexception import UntrustedIdentityException
class SessionBuilderTest(unittest.TestCase):
    ALICE_RECIPIENT_ID = 5
    BOB_RECIPIENT_ID   = 2

    def test_basicPreKeyV2(self):
        aliceStore = InMemoryAxolotlStore()
        aliceSessionBuilder = SessionBuilder(aliceStore, aliceStore, aliceStore, aliceStore, self.__class__.BOB_RECIPIENT_ID, 1)

        bobStore      = InMemoryAxolotlStore()
        bobPreKeyPair = Curve.generateKeyPair()
        bobPreKey     = PreKeyBundle(bobStore.getLocalRegistrationId(), 1,
                                                  31337, bobPreKeyPair.getPublicKey(),
                                                  0, None, None,
                                                  bobStore.getIdentityKeyPair().getPublicKey())

        aliceSessionBuilder.processPreKeyBundle(bobPreKey)

        self.assertTrue(aliceStore.containsSession(self.__class__.BOB_RECIPIENT_ID, 1))
        self.assertEqual(aliceStore.loadSession(self.__class__.BOB_RECIPIENT_ID, 1).getSessionState().getSessionVersion(), 2)

        originalMessage    = "L'homme est condamné à être libre"
        aliceSessionCipher = SessionCipher(aliceStore, aliceStore, aliceStore, aliceStore, self.__class__.BOB_RECIPIENT_ID, 1)
        outgoingMessage    = aliceSessionCipher.encrypt(originalMessage)

        self.assertTrue(outgoingMessage.getType() == CiphertextMessage.PREKEY_TYPE)

        incomingMessage = PreKeyWhisperMessage(serialized=outgoingMessage.serialize())
        bobStore.storePreKey(31337, PreKeyRecord(bobPreKey.getPreKeyId(), bobPreKeyPair))

        bobSessionCipher = SessionCipher(bobStore, bobStore, bobStore, bobStore, self.__class__.ALICE_RECIPIENT_ID, 1)
        plaintext        = bobSessionCipher.decryptPkmsg(incomingMessage)

        self.assertTrue(bobStore.containsSession(self.__class__.ALICE_RECIPIENT_ID, 1))
        self.assertTrue(bobStore.loadSession(self.__class__.ALICE_RECIPIENT_ID, 1).getSessionState().getSessionVersion() == 2)
        self.assertEqual(originalMessage, plaintext)


        bobOutgoingMessage = bobSessionCipher.encrypt(originalMessage)
        self.assertTrue(bobOutgoingMessage.getType() == CiphertextMessage.WHISPER_TYPE)

        alicePlaintext = aliceSessionCipher.decryptMsg(bobOutgoingMessage)
        self.assertEqual(alicePlaintext, originalMessage)

        self.runInteraction(aliceStore, bobStore)

        aliceStore          = InMemoryAxolotlStore()
        aliceSessionBuilder = SessionBuilder(aliceStore, aliceStore, aliceStore, aliceStore, self.__class__.BOB_RECIPIENT_ID, 1)
        aliceSessionCipher  = SessionCipher(aliceStore, aliceStore, aliceStore, aliceStore, self.__class__.BOB_RECIPIENT_ID, 1)

        bobPreKeyPair = Curve.generateKeyPair()
        bobPreKey = PreKeyBundle(bobStore.getLocalRegistrationId(),
                                 1, 31338, bobPreKeyPair.getPublicKey(),
                                 0, None, None, bobStore.getIdentityKeyPair().getPublicKey())

        bobStore.storePreKey(31338, PreKeyRecord(bobPreKey.getPreKeyId(), bobPreKeyPair))
        aliceSessionBuilder.processPreKeyBundle(bobPreKey)

        outgoingMessage = aliceSessionCipher.encrypt(originalMessage)
        try :
            bobSessionCipher.decryptPkmsg(PreKeyWhisperMessage(serialized=outgoingMessage.serialize()))
            raise AssertionError("shouldn't be trusted!")
        except Exception:
            bobStore.saveIdentity(self.__class__.ALICE_RECIPIENT_ID, PreKeyWhisperMessage(serialized=outgoingMessage.serialize()).getIdentityKey())

        plaintext = bobSessionCipher.decryptPkmsg(PreKeyWhisperMessage(serialized=outgoingMessage.serialize()))
        self.assertEqual(plaintext, originalMessage)

        bobPreKey = PreKeyBundle(bobStore.getLocalRegistrationId(), 1,
                                 31337, Curve.generateKeyPair().getPublicKey(),
                                 0, None, None,
                                 aliceStore.getIdentityKeyPair().getPublicKey())
        try:
            aliceSessionBuilder.processPreKeyBundle(bobPreKey)
            raise AssertionError("shouldn't be trusted")
        except Exception:
            #good
            pass

        return


    def test_basicPreKeyV3(self):
        aliceStore = InMemoryAxolotlStore()
        aliceSessionBuilder = SessionBuilder(aliceStore, aliceStore, aliceStore, aliceStore, self.__class__.BOB_RECIPIENT_ID, 1)

        bobStore =   InMemoryAxolotlStore()
        bobPreKeyPair = Curve.generateKeyPair()
        bobSignedPreKeyPair = Curve.generateKeyPair()
        bobSignedPreKeySignature = Curve.calculateSignature(bobStore.getIdentityKeyPair().getPrivateKey(),
                                                                           bobSignedPreKeyPair.getPublicKey().serialize())

        bobPreKey = PreKeyBundle(bobStore.getLocalRegistrationId(), 1,
                                              31337, bobPreKeyPair.getPublicKey(),
                                              22, bobSignedPreKeyPair.getPublicKey(),
                                              bobSignedPreKeySignature,
                                              bobStore.getIdentityKeyPair().getPublicKey())

        aliceSessionBuilder.processPreKeyBundle(bobPreKey)
        self.assertTrue(aliceStore.containsSession(self.__class__.BOB_RECIPIENT_ID, 1))
        self.assertTrue(aliceStore.loadSession(self.__class__.BOB_RECIPIENT_ID, 1).getSessionState().getSessionVersion() == 3)

        originalMessage    = "L'homme est condamné à être libre"
        aliceSessionCipher = SessionCipher(aliceStore, aliceStore, aliceStore, aliceStore, self.__class__.BOB_RECIPIENT_ID, 1)
        outgoingMessage    = aliceSessionCipher.encrypt(originalMessage)

        self.assertTrue(outgoingMessage.getType() == CiphertextMessage.PREKEY_TYPE)

        incomingMessage = PreKeyWhisperMessage(serialized=outgoingMessage.serialize())
        bobStore.storePreKey(31337, PreKeyRecord(bobPreKey.getPreKeyId(), bobPreKeyPair))
        bobStore.storeSignedPreKey(22, SignedPreKeyRecord(22, int(time.time() * 1000), bobSignedPreKeyPair, bobSignedPreKeySignature))

        bobSessionCipher = SessionCipher(bobStore, bobStore, bobStore, bobStore, self.__class__.ALICE_RECIPIENT_ID, 1)

        plaintext = bobSessionCipher.decryptPkmsg(incomingMessage)
        self.assertEqual(originalMessage, plaintext)
        # @@TODO: in callback assertion
        # self.assertFalse(bobStore.containsSession(self.__class__.ALICE_RECIPIENT_ID, 1))

        self.assertTrue(bobStore.containsSession(self.__class__.ALICE_RECIPIENT_ID, 1))

        self.assertTrue(bobStore.loadSession(self.__class__.ALICE_RECIPIENT_ID, 1).getSessionState().getSessionVersion() == 3)
        self.assertTrue(bobStore.loadSession(self.__class__.ALICE_RECIPIENT_ID, 1).getSessionState().getAliceBaseKey() != None)
        self.assertEqual(originalMessage, plaintext)

        bobOutgoingMessage = bobSessionCipher.encrypt(originalMessage)
        self.assertTrue(bobOutgoingMessage.getType() == CiphertextMessage.WHISPER_TYPE)

        alicePlaintext = aliceSessionCipher.decryptMsg(WhisperMessage(serialized=bobOutgoingMessage.serialize()))
        self.assertEqual(alicePlaintext, originalMessage)

        self.runInteraction(aliceStore, bobStore)

        aliceStore          = InMemoryAxolotlStore()
        aliceSessionBuilder = SessionBuilder(aliceStore, aliceStore, aliceStore, aliceStore, self.__class__.BOB_RECIPIENT_ID, 1)
        aliceSessionCipher  = SessionCipher(aliceStore, aliceStore, aliceStore, aliceStore, self.__class__.BOB_RECIPIENT_ID, 1)

        bobPreKeyPair            = Curve.generateKeyPair()
        bobSignedPreKeyPair      = Curve.generateKeyPair()
        bobSignedPreKeySignature = Curve.calculateSignature(bobStore.getIdentityKeyPair().getPrivateKey(), bobSignedPreKeyPair.getPublicKey().serialize())
        bobPreKey = PreKeyBundle(bobStore.getLocalRegistrationId(),
                                 1, 31338, bobPreKeyPair.getPublicKey(),
                                 23, bobSignedPreKeyPair.getPublicKey(), bobSignedPreKeySignature,
                                 bobStore.getIdentityKeyPair().getPublicKey())

        bobStore.storePreKey(31338, PreKeyRecord(bobPreKey.getPreKeyId(), bobPreKeyPair))
        bobStore.storeSignedPreKey(23, SignedPreKeyRecord(23, int(time.time() * 1000), bobSignedPreKeyPair, bobSignedPreKeySignature))
        aliceSessionBuilder.processPreKeyBundle(bobPreKey)

        outgoingMessage = aliceSessionCipher.encrypt(originalMessage)

        try:
            plaintext = bobSessionCipher.decryptPkmsg(PreKeyWhisperMessage(serialized=outgoingMessage))
            raise AssertionError("shouldn't be trusted!")
        except Exception:
            bobStore.saveIdentity(self.__class__.ALICE_RECIPIENT_ID, PreKeyWhisperMessage(serialized=outgoingMessage.serialize()).getIdentityKey())

        plaintext = bobSessionCipher.decryptPkmsg(PreKeyWhisperMessage(serialized=outgoingMessage.serialize()))
        self.assertEqual(plaintext, originalMessage)


        bobPreKey = PreKeyBundle(bobStore.getLocalRegistrationId(), 1,
                                 31337, Curve.generateKeyPair().getPublicKey(),
                                 23, bobSignedPreKeyPair.getPublicKey(), bobSignedPreKeySignature,
                                 aliceStore.getIdentityKeyPair().getPublicKey())
        try:
            aliceSessionBuilder.process(bobPreKey)
            raise AssertionError("shouldn't be trusted!")
        except Exception:
            #good
            pass

    def test_badSignedPreKeySignature(self):
        aliceStore          = InMemoryAxolotlStore()
        aliceSessionBuilder = SessionBuilder(aliceStore, aliceStore, aliceStore, aliceStore,
                                             self.__class__.BOB_RECIPIENT_ID, 1)

        bobIdentityKeyStore = InMemoryIdentityKeyStore()

        bobPreKeyPair            = Curve.generateKeyPair()
        bobSignedPreKeyPair      = Curve.generateKeyPair()
        bobSignedPreKeySignature = Curve.calculateSignature(bobIdentityKeyStore.getIdentityKeyPair().getPrivateKey(),
                                                                  bobSignedPreKeyPair.getPublicKey().serialize())

        for i in range(0, len(bobSignedPreKeySignature) * 8):
            modifiedSignature = bytearray(bobSignedPreKeySignature[:])
            modifiedSignature[int(i/8)] ^= 0x01 << (i % 8)

            bobPreKey = PreKeyBundle(bobIdentityKeyStore.getLocalRegistrationId(), 1,
                                                31337, bobPreKeyPair.getPublicKey(),
                                                22, bobSignedPreKeyPair.getPublicKey(), modifiedSignature,
                                                bobIdentityKeyStore.getIdentityKeyPair().getPublicKey())

            try:
                aliceSessionBuilder.processPreKeyBundle(bobPreKey)
            except Exception:
                pass
                #good
        bobPreKey = PreKeyBundle(bobIdentityKeyStore.getLocalRegistrationId(), 1,
                                              31337, bobPreKeyPair.getPublicKey(),
                                              22, bobSignedPreKeyPair.getPublicKey(), bobSignedPreKeySignature,
                                              bobIdentityKeyStore.getIdentityKeyPair().getPublicKey())

        aliceSessionBuilder.processPreKeyBundle(bobPreKey)


    def test_basicKeyExchange(self):
        aliceStore          = InMemoryAxolotlStore()
        aliceSessionBuilder = SessionBuilder(aliceStore, aliceStore, aliceStore, aliceStore, self.__class__.BOB_RECIPIENT_ID, 1)

        bobStore          = InMemoryAxolotlStore()
        bobSessionBuilder = SessionBuilder(bobStore, bobStore, bobStore, bobStore, self.__class__.ALICE_RECIPIENT_ID, 1)

        aliceKeyExchangeMessage      = aliceSessionBuilder.processInitKeyExchangeMessage()
        self.assertTrue(aliceKeyExchangeMessage != None)

        aliceKeyExchangeMessageBytes = aliceKeyExchangeMessage.serialize()
        bobKeyExchangeMessage        = bobSessionBuilder.processKeyExchangeMessage(KeyExchangeMessage(serialized=aliceKeyExchangeMessageBytes))

        self.assertTrue(bobKeyExchangeMessage != None)

        bobKeyExchangeMessageBytes = bobKeyExchangeMessage.serialize()
        response                   = aliceSessionBuilder.processKeyExchangeMessage(KeyExchangeMessage(serialized=bobKeyExchangeMessageBytes))

        self.assertTrue(response == None)
        self.assertTrue(aliceStore.containsSession(self.__class__.BOB_RECIPIENT_ID, 1))
        self.assertTrue(bobStore.containsSession(self.__class__.ALICE_RECIPIENT_ID, 1))

        self.runInteraction(aliceStore, bobStore)

        aliceStore              = InMemoryAxolotlStore()
        aliceSessionBuilder     = SessionBuilder(aliceStore, aliceStore, aliceStore, aliceStore, self.__class__.BOB_RECIPIENT_ID, 1)
        aliceKeyExchangeMessage = aliceSessionBuilder.processInitKeyExchangeMessage()

        try:
            bobKeyExchangeMessage = bobSessionBuilder.processKeyExchangeMessage(aliceKeyExchangeMessage)
            raise AssertionError("This identity shouldn't be trusted!")
        except UntrustedIdentityException as uie:
            bobStore.saveIdentity(self.__class__.ALICE_RECIPIENT_ID, aliceKeyExchangeMessage.getIdentityKey())
        bobKeyExchangeMessage = bobSessionBuilder.processKeyExchangeMessage(aliceKeyExchangeMessage)

        self.assertTrue(aliceSessionBuilder.processKeyExchangeMessage(bobKeyExchangeMessage) == None)

        self.runInteraction(aliceStore, bobStore)

    def runInteraction(self, aliceStore, bobStore):
        """
        :type aliceStore: AxolotlStore
        :type  bobStore: AxolotlStore
        """

        aliceSessionCipher = SessionCipher(aliceStore, aliceStore, aliceStore, aliceStore, self.__class__.BOB_RECIPIENT_ID, 1)
        bobSessionCipher   = SessionCipher(bobStore, bobStore, bobStore, bobStore, self.__class__.ALICE_RECIPIENT_ID, 1)

        originalMessage = "smert ze smert"
        aliceMessage = aliceSessionCipher.encrypt(originalMessage)

        self.assertTrue(aliceMessage.getType() == CiphertextMessage.WHISPER_TYPE)

        plaintext = bobSessionCipher.decryptMsg(WhisperMessage(serialized=aliceMessage.serialize()))
        self.assertEqual(plaintext, originalMessage)

        bobMessage = bobSessionCipher.encrypt(originalMessage)

        self.assertTrue(bobMessage.getType() == CiphertextMessage.WHISPER_TYPE)

        plaintext = aliceSessionCipher.decryptMsg(WhisperMessage(serialized=bobMessage.serialize()))
        self.assertEqual(plaintext, originalMessage)

        for i in range(0, 10):
            loopingMessage = "What do we mean by saying that existence precedes essence? " \
                             "We mean that man first of all exists, encounters himself, " \
                             "surges up in the world--and defines himself aftward. %s" % i
            aliceLoopingMessage = aliceSessionCipher.encrypt(loopingMessage)
            loopingPlaintext = bobSessionCipher.decryptMsg(WhisperMessage(serialized=aliceLoopingMessage.serialize()))
            self.assertEqual(loopingPlaintext, loopingMessage)


        for i in range(0, 10):
            loopingMessage = "What do we mean by saying that existence precedes essence? " \
                 "We mean that man first of all exists, encounters himself, " \
                 "surges up in the world--and defines himself aftward. %s" % i
            bobLoopingMessage = bobSessionCipher.encrypt(loopingMessage)

            loopingPlaintext = aliceSessionCipher.decryptMsg(WhisperMessage(serialized=bobLoopingMessage.serialize()))
            self.assertEqual(loopingPlaintext, loopingMessage)

        aliceOutOfOrderMessages = []

        for i in range(0, 10):
            loopingMessage = "What do we mean by saying that existence precedes essence? " \
                 "We mean that man first of all exists, encounters himself, " \
                 "surges up in the world--and defines himself aftward. %s" % i
            aliceLoopingMessage = aliceSessionCipher.encrypt(loopingMessage)
            aliceOutOfOrderMessages.append((loopingMessage, aliceLoopingMessage))

        for i in range(0, 10):
            loopingMessage = "What do we mean by saying that existence precedes essence? " \
                 "We mean that man first of all exists, encounters himself, " \
                 "surges up in the world--and defines himself aftward. %s" % i
            aliceLoopingMessage = aliceSessionCipher.encrypt(loopingMessage)
            loopingPlaintext = bobSessionCipher.decryptMsg(WhisperMessage(serialized=aliceLoopingMessage.serialize()))
            self.assertEqual(loopingPlaintext, loopingMessage)

        for i in range(0, 10):
            loopingMessage = "You can only desire based on what you know: %s" % i
            bobLoopingMessage = bobSessionCipher.encrypt(loopingMessage)

            loopingPlaintext = aliceSessionCipher.decryptMsg(WhisperMessage(serialized=bobLoopingMessage.serialize()))
            self.assertEqual(loopingPlaintext, loopingMessage)

        for aliceOutOfOrderMessage in aliceOutOfOrderMessages:
            outOfOrderPlaintext = bobSessionCipher.decryptMsg(WhisperMessage(serialized=aliceOutOfOrderMessage[1].serialize()))
            self.assertEqual(outOfOrderPlaintext, aliceOutOfOrderMessage[0])



from axolotl.protocol.senderkeydistributionmessage import SenderKeyDistributionMessage
class GroupSessionBuilder:
    def __init__(self, senderKeyStore):
        self.senderKeyStore = senderKeyStore

    def processSender(self, sender, senderKeyDistributionMessage):
        """
        :type sender: str
        :type senderKeyDistributionMessage: SenderKeyDistributionMessage
        """
        senderKeyRecord = self.senderKeyStore.loadSenderKey(sender)
        senderKeyRecord.addSenderKeyState(senderKeyDistributionMessage.getId(),
                                        senderKeyDistributionMessage.getIteration(),
                                        senderKeyDistributionMessage.getChainKey(),
                                        senderKeyDistributionMessage.getSignatureKey())
        self.senderKeyStore.storeSenderKey(sender, senderKeyRecord)


    def process(self, groupId, keyId, iteration, chainKey, signatureKey):
        """
        :type groupId: str
        :type keyId: int
        :type iteration: int
        :type chainKey: bytearray
        :type signatureKey: ECKeyPair
        """
        senderKeyRecord = self.senderKeyStore.loadSenderKey(groupId)
        senderKeyRecord.setSenderKeyState(keyId, iteration, chainKey, signatureKey)
        self.senderKeyStore.storeSenderKey(groupId, senderKeyRecord)

        return SenderKeyDistributionMessage(keyId, iteration, chainKey, signatureKey.getPublicKey())



class ECKeyPair():

    def __init__(self, publicKey, privateKey):
        """
        :type publicKey: ECPublicKey
        :type privateKey: ECPrivateKey
        """
        self.publicKey = publicKey
        self.privateKey = privateKey

    def getPrivateKey(self):
        return self.privateKey

    def getPublicKey(self):
        return self.publicKey




class SymmetricAxolotlParameters:
    def __init__(self, ourBaseKey, ourRatchetKey,
                 ourIdentityKey, theirBaseKey,
                 theirRatchetKey, theirIdentityKey):
        """
        :type ourBaseKey: ECKeyPair
        :type ourRatchetKey: ECKeyPair
        :type ourIdentityKey: IdentityKeyPair
        :type theirBaseKey: ECPublicKey
        :type theirRatchetKey: ECKeyPair
        :type theirIdentityKey: IdentityKey
        """

        self.ourBaseKey = ourBaseKey
        self.ourIdentityKey = ourIdentityKey
        self.ourRatchetKey = ourRatchetKey
        self.theirRatchetKey = theirRatchetKey
        self.theirIdentityKey = theirIdentityKey
        self.theirBaseKey = theirBaseKey

        if ourBaseKey is None or ourIdentityKey is None or ourRatchetKey is None\
            or theirRatchetKey is None or theirIdentityKey is None or theirBaseKey is None:
            raise ValueError("Null value!")

    def getOurBaseKey(self):
        return self.ourBaseKey

    def getOurIdentityKey(self):
        return self.ourIdentityKey

    def getTheirRatchetKey(self):
        return self.theirRatchetKey

    def getTheirIdentityKey(self):
        return self.theirIdentityKey

    def getTheirBaseKey(self):
        return self.theirBaseKey

    def getOurRatchetKey(self):
        return self.ourRatchetKey

    @staticmethod
    def newBuilder():
        return SymmetricAxolotlParameters.Builder()

    class Builder:
        def __init__(self):
            self.ourIdentityKey = None
            self.ourBaseKey = None
            self.ourRatchetKey = None
            self.theirRatchetKey = None
            self.theirIdentityKey = None
            self.theirBaseKey = None

        def setOurIdentityKey(self, ourIdentityKey):
            self.ourIdentityKey = ourIdentityKey
            return self

        def setOurBaseKey(self, ourBaseKey):
            self.ourBaseKey = ourBaseKey
            return self

        def setOurRatchetKey(self, ourRatchetKey):
            self.ourRatchetKey = ourRatchetKey
            return self

        def setTheirRatchetKey(self, theirRatchetKey):
            self.theirRatchetKey = theirRatchetKey
            return self



        def setTheirIdentityKey(self, theirIdentityKey):
            self.theirIdentityKey = theirIdentityKey
            return self

        def setTheirBaseKey(self, theirBaseKey):
            self.theirBaseKey = theirBaseKey
            return self

        def create(self):
            return SymmetricAxolotlParameters(self.ourBaseKey, self.ourRatchetKey, self.ourIdentityKey,
                                        self.theirBaseKey, self.theirRatchetKey, self.theirIdentityKey)


import abc
class SenderKeyStore(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def storeSenderKey(self, senderKeyId, senderKeyRecord):
        """
        :type senderKeyId: str
        :type senderKeyRecord: SenderKeyRecord
        """

    @abc.abstractmethod
    def loadSenderKey(self, senderKeyId):
        """
        :type senderKeyId: str
        """


# coding=utf-8
import unittest
from axolotl.state.sessionrecord import SessionRecord
from axolotl.ecc.curve import Curve
from axolotl.identitykeypair import IdentityKeyPair, IdentityKey
from axolotl.ratchet.aliceaxolotlparameters import AliceAxolotlParameters
from axolotl.ratchet.bobaxolotlparamaters import BobAxolotlParameters
from axolotl.ratchet.ratchetingsession import RatchetingSession
from axolotl.tests.inmemoryaxolotlstore import InMemoryAxolotlStore
from axolotl.sessioncipher import SessionCipher
from axolotl.protocol.whispermessage import WhisperMessage
import time
from random import shuffle
class SessionCipherTest(unittest.TestCase):

    def test_basicSessionV2(self):
        aliceSessionRecord = SessionRecord()
        bobSessionRecord = SessionRecord()
        self.initializeSessionsV2(aliceSessionRecord.getSessionState(), bobSessionRecord.getSessionState())
        self.runInteraction(aliceSessionRecord, bobSessionRecord)


    def test_basicSessionV3(self):
        aliceSessionRecord = SessionRecord()
        bobSessionRecord = SessionRecord()
        self.initializeSessionsV3(aliceSessionRecord.getSessionState(), bobSessionRecord.getSessionState())
        self.runInteraction(aliceSessionRecord, bobSessionRecord)

    def runInteraction(self, aliceSessionRecord, bobSessionRecord):
        aliceStore = InMemoryAxolotlStore()
        bobStore   = InMemoryAxolotlStore()

        aliceStore.storeSession(2, 1, aliceSessionRecord)
        bobStore.storeSession(3, 1, bobSessionRecord)

        aliceCipher    = SessionCipher(aliceStore, aliceStore, aliceStore, aliceStore, 2, 1)
        bobCipher      = SessionCipher(bobStore, bobStore, bobStore, bobStore, 3, 1)

        alicePlaintext = "This is a plaintext message."
        message        = aliceCipher.encrypt(alicePlaintext)
        bobPlaintext   = bobCipher.decryptMsg(WhisperMessage(serialized=message.serialize()))

        self.assertEqual(alicePlaintext, bobPlaintext)

        bobReply      = "This is a message from Bob."
        reply         = bobCipher.encrypt(bobReply)
        receivedReply = aliceCipher.decryptMsg(WhisperMessage( serialized=reply.serialize()))

        self.assertEqual(bobReply, receivedReply)

        aliceCiphertextMessages = []
        alicePlaintextMessages = []

        for i in range(0, 50):
            alicePlaintextMessages.append("смерть за смерть %s" % i)
            aliceCiphertextMessages.append(aliceCipher.encrypt("смерть за смерть %s" % i))

        #shuffle(aliceCiphertextMessages)
        #shuffle(alicePlaintextMessages)


        for i in range(0, int(len(aliceCiphertextMessages)/2)):
            receivedPlaintext = bobCipher.decryptMsg(WhisperMessage(serialized=aliceCiphertextMessages[i].serialize()))
            self.assertEqual(receivedPlaintext, alicePlaintextMessages[i])


        """

    List<CiphertextMessage> bobCiphertextMessages = new ArrayList<>();
    List<byte[]>            bobPlaintextMessages  = new ArrayList<>();

    for (int i=0;i<20;i++) {
      bobPlaintextMessages.add(("смерть за смерть " + i).getBytes());
      bobCiphertextMessages.add(bobCipher.encrypt(("смерть за смерть " + i).getBytes()));
    }

    seed = System.currentTimeMillis();

    Collections.shuffle(bobCiphertextMessages, new Random(seed));
    Collections.shuffle(bobPlaintextMessages, new Random(seed));

    for (int i=0;i<bobCiphertextMessages.size() / 2;i++) {
      byte[] receivedPlaintext = aliceCipher.decrypt(new WhisperMessage(bobCiphertextMessages.get(i).serialize()));
      assertTrue(Arrays.equals(receivedPlaintext, bobPlaintextMessages.get(i)));
    }

    for (int i=aliceCiphertextMessages.size()/2;i<aliceCiphertextMessages.size();i++) {
      byte[] receivedPlaintext = bobCipher.decrypt(new WhisperMessage(aliceCiphertextMessages.get(i).serialize()));
      assertTrue(Arrays.equals(receivedPlaintext, alicePlaintextMessages.get(i)));
    }

    for (int i=bobCiphertextMessages.size() / 2;i<bobCiphertextMessages.size();i++) {
      byte[] receivedPlaintext = aliceCipher.decrypt(new WhisperMessage(bobCiphertextMessages.get(i).serialize()));
      assertTrue(Arrays.equals(receivedPlaintext, bobPlaintextMessages.get(i)));
    }
        """




    def initializeSessionsV2(self, aliceSessionState, bobSessionState):
        aliceIdentityKeyPair = Curve.generateKeyPair()
        aliceIdentityKey     = IdentityKeyPair(IdentityKey(aliceIdentityKeyPair.getPublicKey()),
                                                               aliceIdentityKeyPair.getPrivateKey())
        aliceBaseKey         = Curve.generateKeyPair()
        aliceEphemeralKey    = Curve.generateKeyPair()

        bobIdentityKeyPair   = Curve.generateKeyPair()
        bobIdentityKey       = IdentityKeyPair(IdentityKey(bobIdentityKeyPair.getPublicKey()),
                                                               bobIdentityKeyPair.getPrivateKey())
        bobBaseKey           = Curve.generateKeyPair()
        bobEphemeralKey      = bobBaseKey

        aliceParameters = AliceAxolotlParameters.newBuilder()\
            .setOurIdentityKey(aliceIdentityKey)\
            .setOurBaseKey(aliceBaseKey)\
            .setTheirIdentityKey(bobIdentityKey.getPublicKey())\
            .setTheirSignedPreKey(bobEphemeralKey.getPublicKey())\
            .setTheirRatchetKey(bobEphemeralKey.getPublicKey())\
            .setTheirOneTimePreKey(None)\
            .create()

        bobParameters = BobAxolotlParameters.newBuilder()\
            .setOurIdentityKey(bobIdentityKey)\
            .setOurOneTimePreKey(None)\
            .setOurRatchetKey(bobEphemeralKey)\
            .setOurSignedPreKey(bobBaseKey)\
            .setTheirBaseKey(aliceBaseKey.getPublicKey())\
            .setTheirIdentityKey(aliceIdentityKey.getPublicKey())\
            .create()

        RatchetingSession.initializeSessionAsAlice(aliceSessionState, 2, aliceParameters)
        RatchetingSession.initializeSessionAsBob(bobSessionState, 2, bobParameters)


    def initializeSessionsV3(self, aliceSessionState, bobSessionState):
        aliceIdentityKeyPair = Curve.generateKeyPair()
        aliceIdentityKey     = IdentityKeyPair(IdentityKey(aliceIdentityKeyPair.getPublicKey()),
                                                               aliceIdentityKeyPair.getPrivateKey())
        aliceBaseKey         = Curve.generateKeyPair()
        aliceEphemeralKey    = Curve.generateKeyPair()

        alicePreKey          = aliceBaseKey

        bobIdentityKeyPair   = Curve.generateKeyPair()
        bobIdentityKey       = IdentityKeyPair(IdentityKey(bobIdentityKeyPair.getPublicKey()),
                                                               bobIdentityKeyPair.getPrivateKey())
        bobBaseKey           = Curve.generateKeyPair()
        bobEphemeralKey      = bobBaseKey

        bobPreKey            = Curve.generateKeyPair()

        aliceParameters = AliceAxolotlParameters.newBuilder()\
            .setOurBaseKey(aliceBaseKey)\
            .setOurIdentityKey(aliceIdentityKey)\
            .setTheirOneTimePreKey(None)\
            .setTheirRatchetKey(bobEphemeralKey.getPublicKey())\
            .setTheirSignedPreKey(bobBaseKey.getPublicKey())\
            .setTheirIdentityKey(bobIdentityKey.getPublicKey())\
            .create()

        bobParameters = BobAxolotlParameters.newBuilder()\
            .setOurRatchetKey(bobEphemeralKey)\
            .setOurSignedPreKey(bobBaseKey)\
            .setOurOneTimePreKey(None)\
            .setOurIdentityKey(bobIdentityKey)\
            .setTheirIdentityKey(aliceIdentityKey.getPublicKey())\
            .setTheirBaseKey(aliceBaseKey.getPublicKey())\
            .create()

        RatchetingSession.initializeSessionAsAlice(aliceSessionState, 3, aliceParameters)
        RatchetingSession.initializeSessionAsBob(bobSessionState, 3, bobParameters)



import codecs, sys
decode_hex = codecs.getdecoder("hex_codec")
class HexUtil:
    @staticmethod
    def decodeHex(hexString):
        hexString = hexString.encode() if type(hexString) is str else hexString
        result = decode_hex(hexString)[0]
        if sys.version_info >= (3,0):
            result = result.decode('latin-1')
        return result




from .ciphertextmessage import CiphertextMessage
from ..util.byteutil import ByteUtil
from ..ecc.curve import Curve
from . import whisperprotos
import hmac
import hashlib
from axolotl.legacymessageexception import LegacyMessageException
from axolotl.invalidmessageexception import InvalidMessageException
from axolotl.invalidkeyexception import InvalidKeyException
class WhisperMessage(CiphertextMessage):
    MAC_LENGTH = 8

    def __init__(self, messageVersion = None, macKey = None, ECPublicKey_senderRatchetKey = None,
                 counter = None, previousCounter = None, ciphertext = None,
                 senderIdentityKey = None, receiverIdentityKey = None,
                 serialized = None):

        self.serialized = ""
        if serialized:
            try:
                assert type(serialized) in (str, bytes), "Expected serialized %s, got %s" % (str, type(serialized))
                messageParts = ByteUtil.split(serialized, 1, len(serialized) - 1 - WhisperMessage.MAC_LENGTH,
                                              WhisperMessage.MAC_LENGTH)
                version = messageParts[0][0]
                message = messageParts[1]
                mac = messageParts[2]

                if ByteUtil.highBitsToInt(version) <= self.__class__.UNSUPPORTED_VERSION:
                    raise LegacyMessageException("Legacy message %s" % ByteUtil.highBitsToInt(version))

                if ByteUtil.highBitsToInt(version) > self.__class__.CURRENT_VERSION:
                    raise InvalidMessageException("Unknown version: %s" % ByteUtil.highBitsToInt(version))

                whisperMessage = whisperprotos.WhisperMessage()
                whisperMessage.ParseFromString(message)

                if not whisperMessage.ciphertext or whisperMessage.counter is None or not whisperMessage.ratchetKey:
                    raise InvalidMessageException("Incomplete message")

                self.serialized = serialized
                self.senderRatchetKey = Curve.decodePoint(bytearray(whisperMessage.ratchetKey), 0)
                self.messageVersion = ByteUtil.highBitsToInt(version)
                self.counter = whisperMessage.counter
                self.previousCounter = whisperMessage.previousCounter
                self.ciphertext = whisperMessage.ciphertext
            except InvalidKeyException as e:
                raise InvalidMessageException(e)
        else:
            version  = ByteUtil.intsToByteHighAndLow(messageVersion, self.__class__.CURRENT_VERSION)
            message = whisperprotos.WhisperMessage()
            message.ratchetKey = ECPublicKey_senderRatchetKey.serialize()
            message.counter = counter
            message.previousCounter = previousCounter
            message.ciphertext = ciphertext
            message = message.SerializeToString()
            mac  = self.getMac(messageVersion, senderIdentityKey, receiverIdentityKey, macKey,
                               ByteUtil.combine(version, message))
            self.serialized = bytes(ByteUtil.combine(version, message, mac))
            self.senderRatchetKey = ECPublicKey_senderRatchetKey
            self.counter = counter
            self.previousCounter = previousCounter
            self.ciphertext = ciphertext
            self.messageVersion = messageVersion


    def getSenderRatchetKey(self):
        return self.senderRatchetKey

    def getMessageVersion(self):
        return self.messageVersion

    def getCounter(self):
        return self.counter

    def getBody(self):
        return self.ciphertext

    def verifyMac(self, messageVersion, senderIdentityKey, receiverIdentityKey, macKey):
        parts = ByteUtil.split(self.serialized, len(self.serialized) - self.__class__.MAC_LENGTH, self.__class__.MAC_LENGTH)
        ourMac = self.getMac(messageVersion, senderIdentityKey, receiverIdentityKey, macKey, parts[0])
        theirMac = parts[1]

        if ourMac != theirMac:
            raise InvalidMessageException("Bad Mac!")

    def getMac(self, messageVersion, senderIdentityKey, receiverIdentityKey, macKey, serialized):
        mac = hmac.new(macKey, digestmod=hashlib.sha256)
        if messageVersion >= 3:
            mac.update(senderIdentityKey.getPublicKey().serialize())
            mac.update(receiverIdentityKey.getPublicKey().serialize())

        mac.update(bytes(serialized))
        fullMac = mac.digest()
        return ByteUtil.trim(fullMac, self.__class__.MAC_LENGTH)

    def serialize(self):
        return self.serialized

    def getType(self):
        return CiphertextMessage.WHISPER_TYPE

    def isLegacy(self, message):
        return message is not None and len(message) >= 1 and ByteUtil.highBitsToInt(message[0]) <= CiphertextMessage.UNSUPPORTED_VERSION



import logging
from .ratchet.ratchetingsession import RatchetingSession
from .ecc.curve import Curve
from .util.medium import Medium
from .ratchet.aliceaxolotlparameters import AliceAxolotlParameters
from .ratchet.bobaxolotlparamaters import BobAxolotlParameters
from axolotl.invalidkeyexception import InvalidKeyException
from axolotl.invalidkeyidexception import InvalidKeyIdException
from axolotl.untrustedidentityexception import UntrustedIdentityException
from axolotl.protocol.keyexchangemessage import KeyExchangeMessage
from axolotl.ratchet.symmetricaxolotlparameters import SymmetricAxolotlParameters
from axolotl.protocol.ciphertextmessage import CiphertextMessage
from axolotl.statekeyexchangeexception import StaleKeyExchangeException
from axolotl.util.keyhelper import KeyHelper
logger = logging.getLogger(__name__)
class SessionBuilder:
    def __init__(self, sessionStore, preKeyStore, signedPreKeyStore, identityKeyStore, recepientId, deviceId):
        self.sessionStore = sessionStore
        self.preKeyStore = preKeyStore
        self.signedPreKeyStore = signedPreKeyStore
        self.identityKeyStore = identityKeyStore
        self.recipientId = recepientId
        self.deviceId = deviceId

    def process(self, sessionRecord, message):
        """
        :param sessionRecord:
        :param message:
        :type message: PreKeyWhisperMessage
        """

        messageVersion = message.getMessageVersion()
        theirIdentityKey = message.getIdentityKey()

        unsignedPreKeyId = None

        if not self.identityKeyStore.isTrustedIdentity(self.recipientId, theirIdentityKey):
            raise UntrustedIdentityException("Untrusted identity!!")

        if messageVersion == 2:
            unsignedPreKeyId = self.processV2(sessionRecord, message)
        elif messageVersion == 3:
            unsignedPreKeyId = self.processV3(sessionRecord, message)
        else:
            raise AssertionError("Unkown version %s" % messageVersion)

        self.identityKeyStore.saveIdentity(self.recipientId, theirIdentityKey)

        return unsignedPreKeyId


    def processV2(self, sessionRecord, message):
        """
        :type sessionRecord: SessionRecord
        :type message: PreKeyWhisperMessage
        """

        if message.getPreKeyId() is None:
            raise InvalidKeyIdException("V2 message requires one time prekey id!")

        if not self.preKeyStore.containsPreKey(message.getPreKeyId()) and\
            self.sessionStore.containsSession(self.recipientId, self.deviceId):
            logging.warn("We've already processed the prekey part of this V2 session, letting bundled message fall through...")
            return None

        ourPreKey = self.preKeyStore.loadPreKey(message.getPreKeyId()).getKeyPair()

        parameters = BobAxolotlParameters.newBuilder()

        parameters.setOurIdentityKey(self.identityKeyStore.getIdentityKeyPair())\
              .setOurSignedPreKey(ourPreKey)\
              .setOurRatchetKey(ourPreKey)\
              .setOurOneTimePreKey(None)\
              .setTheirIdentityKey(message.getIdentityKey())\
              .setTheirBaseKey(message.getBaseKey())

        if not sessionRecord.isFresh():  sessionRecord.archiveCurrentState()

        RatchetingSession.initializeSessionAsBob(sessionRecord.getSessionState(), message.getMessageVersion(), parameters.create())

        sessionRecord.getSessionState().setLocalRegistrationId(self.identityKeyStore.getLocalRegistrationId())
        sessionRecord.getSessionState().setRemoteRegistrationId(message.getRegistrationId())
        sessionRecord.getSessionState().setAliceBaseKey(message.getBaseKey().serialize())

        if message.getPreKeyId() != Medium.MAX_VALUE:
            return message.getPreKeyId()
        else:
            return None


    def processV3(self, sessionRecord, message):
        """

        :param sessionRecord:
        :param message:
        :type message: PreKeyWhisperMessage
        :return:
        """

        if sessionRecord.hasSessionState(message.getMessageVersion(), message.getBaseKey().serialize()):
            logger.warn("We've already setup a session for this V3 message, letting bundled message fall through...")
            return None

        ourSignedPreKey = self.signedPreKeyStore.loadSignedPreKey(message.getSignedPreKeyId()).getKeyPair()
        parameters = BobAxolotlParameters.newBuilder()
        parameters.setTheirBaseKey(message.getBaseKey())\
            .setTheirIdentityKey(message.getIdentityKey())\
            .setOurIdentityKey(self.identityKeyStore.getIdentityKeyPair())\
            .setOurSignedPreKey(ourSignedPreKey)\
            .setOurRatchetKey(ourSignedPreKey)

        if message.getPreKeyId() is not None:
            parameters.setOurOneTimePreKey(self.preKeyStore.loadPreKey(message.getPreKeyId()).getKeyPair())
        else:
            parameters.setOurOneTimePreKey(None)

        if not sessionRecord.isFresh():
            sessionRecord.archiveCurrentState()

        RatchetingSession.initializeSessionAsBob(sessionRecord.getSessionState(), message.getMessageVersion(), parameters.create())
        sessionRecord.getSessionState().setLocalRegistrationId(self.identityKeyStore.getLocalRegistrationId())
        sessionRecord.getSessionState().setRemoteRegistrationId(message.getRegistrationId())
        sessionRecord.getSessionState().setAliceBaseKey(message.getBaseKey().serialize())

        if message.getPreKeyId() is not None and message.getPreKeyId() != Medium.MAX_VALUE:
            return message.getPreKeyId()
        else:
            return None


    def processPreKeyBundle(self, preKey):
        """
        :type preKey: PreKeyBundle
        """
        if not self.identityKeyStore.isTrustedIdentity(self.recipientId, preKey.getIdentityKey()):
            raise UntrustedIdentityException()

        if preKey.getSignedPreKey() is not None and\
            not Curve.verifySignature(preKey.getIdentityKey().getPublicKey(),
                                      preKey.getSignedPreKey().serialize(),
                                      preKey.getSignedPreKeySignature()) :
            raise InvalidKeyException("Invalid signature on device key!")

        if preKey.getSignedPreKey() is None and preKey.getPreKey() is None:
            raise InvalidKeyException("Both signed and unsigned prekeys are absent!")

        supportsV3 = preKey.getSignedPreKey() is not None
        sessionRecord        = self.sessionStore.loadSession(self.recipientId, self.deviceId)
        ourBaseKey           = Curve.generateKeyPair()
        theirSignedPreKey    = preKey.getSignedPreKey() if supportsV3 else preKey.getPreKey()
        theirOneTimePreKey   = preKey.getPreKey()
        theirOneTimePreKeyId = preKey.getPreKeyId() if theirOneTimePreKey is not None else None

        parameters = AliceAxolotlParameters.newBuilder()

        parameters.setOurBaseKey(ourBaseKey)\
                .setOurIdentityKey(self.identityKeyStore.getIdentityKeyPair())\
                .setTheirIdentityKey(preKey.getIdentityKey())\
                .setTheirSignedPreKey(theirSignedPreKey)\
                .setTheirRatchetKey(theirSignedPreKey)\
                .setTheirOneTimePreKey(theirOneTimePreKey if supportsV3 else None)

        if not sessionRecord.isFresh(): sessionRecord.archiveCurrentState();

        RatchetingSession.initializeSessionAsAlice(sessionRecord.getSessionState(),
                                                   3 if supportsV3 else 2,
                                                   parameters.create())

        sessionRecord.getSessionState().setUnacknowledgedPreKeyMessage(theirOneTimePreKeyId, preKey.getSignedPreKeyId(), ourBaseKey.getPublicKey())
        sessionRecord.getSessionState().setLocalRegistrationId(self.identityKeyStore.getLocalRegistrationId())
        sessionRecord.getSessionState().setRemoteRegistrationId(preKey.getRegistrationId())
        sessionRecord.getSessionState().setAliceBaseKey(ourBaseKey.getPublicKey().serialize())
        self.sessionStore.storeSession(self.recipientId, self.deviceId, sessionRecord)
        self.identityKeyStore.saveIdentity(self.recipientId, preKey.getIdentityKey())

    def processKeyExchangeMessage(self, keyExchangeMessage):

        if not self.identityKeyStore.isTrustedIdentity(self.recipientId, keyExchangeMessage.getIdentityKey()):
            raise UntrustedIdentityException()

        responseMessage = None

        if keyExchangeMessage.isInitiate(): responseMessage = self.processInitiate(keyExchangeMessage)
        else: self.processResponse(keyExchangeMessage)

        return responseMessage

    def processInitiate(self, keyExchangeMessage):
        flags = KeyExchangeMessage.RESPONSE_FLAG
        sessionRecord = self.sessionStore.loadSession(self.recipientId, self.deviceId)

        if keyExchangeMessage.getVersion() >= 3 and not Curve.verifySignature(
                keyExchangeMessage.getIdentityKey().getPublicKey(),
                keyExchangeMessage.getBaseKey().serialize(),
                keyExchangeMessage.getBaseKeySignature()):
            raise InvalidKeyException("Bad signature!")

        builder = SymmetricAxolotlParameters.newBuilder()
        if not sessionRecord.getSessionState().hasPendingKeyExchange():
            builder.setOurIdentityKey(self.identityKeyStore.getIdentityKeyPair())\
                .setOurBaseKey(Curve.generateKeyPair())\
                .setOurRatchetKey(Curve.generateKeyPair())
        else:
            builder.setOurIdentityKey(sessionRecord.getSessionState().getPendingKeyExchangeIdentityKey())\
                .setOurBaseKey(sessionRecord.getSessionState().getPendingKeyExchangeBaseKey())\
                .setOurRatchetKey(sessionRecord.getSessionState().getPendingKeyExchangeRatchetKey())
            flags |= KeyExchangeMessage.SIMULTAENOUS_INITIATE_FLAG


        builder.setTheirBaseKey(keyExchangeMessage.getBaseKey())\
            .setTheirRatchetKey(keyExchangeMessage.getRatchetKey())\
            .setTheirIdentityKey(keyExchangeMessage.getIdentityKey())

        parameters = builder.create()

        if not sessionRecord.isFresh(): sessionRecord.archiveCurrentState()

        RatchetingSession.initializeSession(sessionRecord.getSessionState(),
                                        min(keyExchangeMessage.getMaxVersion(), CiphertextMessage.CURRENT_VERSION),
                                        parameters)

        self.sessionStore.storeSession(self.recipientId, self.deviceId, sessionRecord)
        self.identityKeyStore.saveIdentity(self.recipientId, keyExchangeMessage.getIdentityKey())

        baseKeySignature = Curve.calculateSignature(parameters.getOurIdentityKey().getPrivateKey(),
                                                       parameters.getOurBaseKey().getPublicKey().serialize())

        return KeyExchangeMessage(sessionRecord.getSessionState().getSessionVersion(),
                                  keyExchangeMessage.getSequence(), flags,
                                  parameters.getOurBaseKey().getPublicKey(),
                                  baseKeySignature, parameters.getOurRatchetKey().getPublicKey(),
                                  parameters.getOurIdentityKey().getPublicKey())

    def processResponse(self, keyExchangeMessage):
        sessionRecord                  = self.sessionStore.loadSession(self.recipientId, self.deviceId)
        sessionState                   = sessionRecord.getSessionState()
        hasPendingKeyExchange          = sessionState.hasPendingKeyExchange()
        isSimultaneousInitiateResponse = keyExchangeMessage.isResponseForSimultaneousInitiate()

        if not hasPendingKeyExchange or sessionState.getPendingKeyExchangeSequence() != keyExchangeMessage.getSequence():
            logger.warn("No matching sequence for response. Is simultaneous initiate response: %s" % isSimultaneousInitiateResponse)
            if not isSimultaneousInitiateResponse:
                raise StaleKeyExchangeException()
            else:
                return

        parameters = SymmetricAxolotlParameters.newBuilder()

        parameters.setOurBaseKey(sessionRecord.getSessionState().getPendingKeyExchangeBaseKey())\
            .setOurRatchetKey(sessionRecord.getSessionState().getPendingKeyExchangeRatchetKey())\
            .setOurIdentityKey(sessionRecord.getSessionState().getPendingKeyExchangeIdentityKey())\
            .setTheirBaseKey(keyExchangeMessage.getBaseKey())\
            .setTheirRatchetKey(keyExchangeMessage.getRatchetKey())\
            .setTheirIdentityKey(keyExchangeMessage.getIdentityKey())

        if not sessionRecord.isFresh(): sessionRecord.archiveCurrentState()

        RatchetingSession.initializeSession(sessionRecord.getSessionState(),
                                        min(keyExchangeMessage.getMaxVersion(), CiphertextMessage.CURRENT_VERSION),
                                        parameters.create())

        if sessionRecord.getSessionState().getSessionVersion() >= 3 and not Curve.verifySignature(
                keyExchangeMessage.getIdentityKey().getPublicKey(),
                keyExchangeMessage.getBaseKey().serialize(),
                keyExchangeMessage.getBaseKeySignature()):
            raise InvalidKeyException("Base key signature doesn't match!")


        self.sessionStore.storeSession(self.recipientId, self.deviceId, sessionRecord)
        self.identityKeyStore.saveIdentity(self.recipientId, keyExchangeMessage.getIdentityKey())

    def processInitKeyExchangeMessage(self):
        try:
            sequence         = KeyHelper.getRandomSequence(65534) + 1
            flags            = KeyExchangeMessage.INITIATE_FLAG
            baseKey          = Curve.generateKeyPair()
            ratchetKey       = Curve.generateKeyPair()
            identityKey      = self.identityKeyStore.getIdentityKeyPair()
            baseKeySignature = Curve.calculateSignature(identityKey.getPrivateKey(), baseKey.getPublicKey().serialize())
            sessionRecord    = self.sessionStore.loadSession(self.recipientId, self.deviceId)

            sessionRecord.getSessionState().setPendingKeyExchange(sequence, baseKey, ratchetKey, identityKey)
            self.sessionStore.storeSession(self.recipientId, self.deviceId, sessionRecord)

            return KeyExchangeMessage(2, sequence, flags, baseKey.getPublicKey(), baseKeySignature,
                                      ratchetKey.getPublicKey(), identityKey.getPublicKey())
        except InvalidKeyException as e:
            raise AssertionError(e)


class PreKeyBundle:
    def __init__(self, registrationId, deviceId, preKeyId, ECPublicKey_preKeyPublic,
                 signedPreKeyId, ECPublicKey_signedPreKeyPublic, signedPreKeySignature,
                 identityKey):
        self.registrationId = registrationId
        self.deviceId = deviceId
        self.preKeyId = preKeyId
        self.preKeyPublic = ECPublicKey_preKeyPublic
        self.signedPreKeyId = signedPreKeyId
        self.signedPreKeyPublic = ECPublicKey_signedPreKeyPublic
        self.signedPreKeySignature = signedPreKeySignature
        self.identityKey = identityKey

    def getDeviceId(self):
        return self.deviceId

    def getPreKeyId(self):
        return self.preKeyId

    def getPreKey(self):
        return self.preKeyPublic

    def getSignedPreKeyId(self):
        return self.signedPreKeyId

    def getSignedPreKey(self):
        return self.signedPreKeyPublic

    def getSignedPreKeySignature(self):
        return self.signedPreKeySignature

    def getIdentityKey(self):
        return self.identityKey

    def getRegistrationId(self):
        return self.registrationId


class MessageKeys:
    def __init__(self, cipherKey, macKey, iv, counter):
        self.cipherKey = cipherKey
        self.macKey = macKey
        self.iv = iv
        self.counter = counter

    def getCipherKey(self):
        return self.cipherKey

    def getMacKey(self):
        return self.macKey

    def getIv(self):
        return self.iv

    def getCounter(self):
        return self.counter


import unittest
import binascii
from axolotl.protocol.whispermessage import WhisperMessage
from axolotl.protocol.prekeywhispermessage import PreKeyWhisperMessage
class TestCoderus(unittest.TestCase):
    def test_whisper(self):
        serialized = "230a21056013e4e2a2f72d131b5b10f760b46f9c087fbb33ef67baa26ed9b17e68a8963a10"
        serialized = "330a21053ac1b1b52beb88f4cffcbc92c39760b593da0c691d82133a0d881e9ee9a2bb621031180022301d9b8af1626a2e379baabb0ec1b299c2965aa40d4c0408b6c2001d60ac317dd880bedc185b51753cc5f5bfdd9c2336c055afcaadaa673f2a"
        serialized = binascii.unhexlify(serialized)
        # whisper = WhisperMessage(serialized = serialized)
        # prewhisperSerialized = "2308e9f401122105f36285d7e388e014b15dbc035202c0d6dcd7ece4e811d374cc5c91462c4889341a2105ce34be348f012e18d25914131713588f15f3f0c3ad56e72a7c1d4910d542a9292225230a21059d5ee0bb2ba7843e06c08dcb2f611ab7112ae516c8ae2a31625e0d9ed4baf10e10289192d3b20f3000"
        prewhisperSerialized = "2308e9f401122105db5e6eb1a65595ae56482995a94fb9400af928d4d73bb548ed493c13a305bb3d1a2105f08eef48c14f94c7c5ebc9fb5c7dc6fe867567920c285aa90983220d2509b94c2256230a2105767fc2f37de0b8130e1808cf6b505c636d118adcb99f793d14f5bc263a6ae44610001800222466b9df118ebd28c7064ef7f6c0366205bf56160e6592c6358311944402e3167382f8670298653b59b1cbf76d28fbb4d8fb093000"
        prewhisperSerialized = binascii.unhexlify(prewhisperSerialized)
        p = PreKeyWhisperMessage(serialized=prewhisperSerialized)


import abc
class CiphertextMessage(object):
    __metaclass__ = abc.ABCMeta

    UNSUPPORTED_VERSION = 1
    CURRENT_VERSION     = 3

    WHISPER_TYPE                = 2
    PREKEY_TYPE                 = 3
    SENDERKEY_TYPE              = 4
    SENDERKEY_DISTRIBUTION_TYPE = 5

    #This should be the worst case (worse than V2).  So not always accurate, but good enough for padding.
    ENCRYPTED_MESSAGE_OVERHEAD = 53
    @abc.abstractmethod
    def serialize(self):
        return

    @abc.abstractmethod
    def getType(self):
        return


import abc
class IdentityKeyStore(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def getIdentityKeyPair(self):pass

    @abc.abstractmethod
    def getLocalRegistrationId(self):pass

    @abc.abstractmethod
    def saveIdentity(self, recepientId, identityKey):pass

    @abc.abstractmethod
    def isTrustedIdentity(self, recepientId, identityKey):pass


from axolotl.state.axolotlstore import AxolotlStore
from axolotl.tests.inmemoryidentitykeystore import InMemoryIdentityKeyStore
from axolotl.tests.inmemoryprekeystore import InMemoryPreKeyStore
from axolotl.tests.inmemorysessionstore import InMemorySessionStore
from axolotl.tests.inmemorysignedprekeystore import InMemorySignedPreKeyStore

class InMemoryAxolotlStore(AxolotlStore):
    def __init__(self):
        self.identityKeyStore = InMemoryIdentityKeyStore()
        self.preKeyStore = InMemoryPreKeyStore()
        self.signedPreKeyStore = InMemorySignedPreKeyStore()
        self.sessionStore = InMemorySessionStore()

    def getIdentityKeyPair(self):
        return self.identityKeyStore.getIdentityKeyPair()

    def getLocalRegistrationId(self):
        return self.identityKeyStore.getLocalRegistrationId()

    def saveIdentity(self, recepientId, identityKey):
        self.identityKeyStore.saveIdentity(recepientId, identityKey)

    def isTrustedIdentity(self, recepientId, identityKey):
        return self.identityKeyStore.isTrustedIdentity(recepientId, identityKey)

    def loadPreKey(self, preKeyId):
        return self.preKeyStore.loadPreKey(preKeyId)

    def storePreKey(self, preKeyId, preKeyRecord):
        self.preKeyStore.storePreKey(preKeyId, preKeyRecord)

    def containsPreKey(self, preKeyId):
        return self.preKeyStore.containsPreKey(preKeyId)

    def removePreKey(self, preKeyId):
        self.preKeyStore.removePreKey(preKeyId)

    def loadSession(self, recepientId, deviceId):
        return self.sessionStore.loadSession(recepientId, deviceId)

    def getSubDeviceSessions(self, recepientId):
        return self.sessionStore.getSubDeviceSessions(recepientId)

    def storeSession(self, recepientId, deviceId, sessionRecord):
        self.sessionStore.storeSession(recepientId, deviceId, sessionRecord)

    def containsSession(self, recepientId, deviceId):
        return self.sessionStore.containsSession(recepientId, deviceId)

    def deleteSession(self, recepientId, deviceId):
        self.sessionStore.deleteSession(recepientId, deviceId)

    def deleteAllSessions(self, recepientId):
        self.sessionStore.deleteAllSessions(recepientId)

    def loadSignedPreKey(self, signedPreKeyId):
        return self.signedPreKeyStore.loadSignedPreKey(signedPreKeyId)

    def loadSignedPreKeys(self):
        return self.signedPreKeyStore.loadSignedPreKeys()

    def storeSignedPreKey(self, signedPreKeyId, signedPreKeyRecord):
        self.signedPreKeyStore.storeSignedPreKey(signedPreKeyId, signedPreKeyRecord)

    def containsSignedPreKey(self, signedPreKeyId):
        return self.signedPreKeyStore.containsSignedPreKey(signedPreKeyId)

    def removeSignedPreKey(self, signedPreKeyId):
        return self.signedPreKeyStore.containsSignedPreKey()



from axolotl.state import storageprotos
from axolotl.groups.ratchet.senderchainkey import SenderChainKey
from axolotl.groups.ratchet.sendermessagekey import SenderMessageKey
from axolotl.ecc.curve import Curve
class SenderKeyState:
    def __init__(self, id = None, iteration = None, chainKey = None,
                 signatureKeyPublic = None, signatureKeyPrivate = None
                 , signatureKeyPair = None, senderKeyStateStructure = None):
        """
        :type id: int
        :type iteration: int
        :type chainKey:  bytearray
        :type signatureKeyPublic:  ECPublicKey
        :type signatureKeyPrivate: ECPrivateKey
        :type signatureKeyPair: ECKeyPair
        :type senderKeyStateStructure: SenderKeyStateStructure
        """

        assert (bool(id) and bool(iteration) and bool(chainKey)) or \
               (bool(senderKeyStateStructure) ^ bool(signatureKeyPublic or signatureKeyPair)) or \
               (bool(signatureKeyPublic) ^ bool(signatureKeyPair)), "Missing required arguments"

        if senderKeyStateStructure:
            self.senderKeyStateStructure = senderKeyStateStructure
        else:
            if signatureKeyPair:
                signatureKeyPublic = signatureKeyPair.getPublicKey()
                signatureKeyPrivate = signatureKeyPair.getPrivateKey()


            self.senderKeyStateStructure = storageprotos.SenderKeyStateStructure()
            senderChainKeyStructure = self.senderKeyStateStructure.SenderChainKey()
            senderChainKeyStructure.iteration = iteration
            senderChainKeyStructure.seed = chainKey

            signingKeyStructure = self.senderKeyStateStructure.SenderSigningKey()
            signingKeyStructure.public = signatureKeyPublic.serialize()

            if signatureKeyPrivate:
                signingKeyStructure.private = signatureKeyPrivate.serialize()

            self.senderKeyStateStructure.senderKeyId = id
            self.senderChainKey = senderChainKeyStructure
            self.senderKeyStateStructure.senderSigningKey.CopyFrom(signingKeyStructure)


    def getKeyId(self):
        return self.senderKeyStateStructure.senderKeyId

    def getSenderChainKey(self):
        return SenderChainKey(self.senderKeyStateStructure.senderChainKey.iteration,
                              bytearray(self.senderKeyStateStructure.senderChainKey.seed))


    def setSenderChainKey(self, chainKey):
        self.senderKeyStateStructure.senderChainKey.iteration = chainKey.getIteration()
        self.senderKeyStateStructure.senderChainKey.seed = chainKey.getSeed()

    def getSigningKeyPublic(self):
        return Curve.decodePoint(bytearray(self.senderKeyStateStructure.senderSigningKey.public), 0)

    def getSigningKeyPrivate(self):
        return Curve.decodePrivatePoint(self.senderKeyStateStructure.senderSigningKey.private)


    def hasSenderMessageKey(self, iteration):
        for senderMessageKey in self.senderKeyStateStructure.senderMessageKeys:
            if senderMessageKey.iteration == iteration:
                return True

        return False


    def addSenderMessageKey(self, senderMessageKey):
        smk = self.senderKeyStateStructure.SenderMessageKey()
        smk.iteration = senderMessageKey.iteration
        smk.seed = senderMessageKey.seed
        self.senderKeyStateStructure.senderMessageKeys.extend([smk])

    def removeSenderMessageKey(self, iteration):
        keys = self.senderKeyStateStructure.senderMessageKeys
        result = None

        for i in range(0, len(keys)):
            senderMessageKey = keys[i]
            if senderMessageKey.iteration == iteration:
                result = senderMessageKey
                del keys[i]
                break

        if result is not None:
            return SenderMessageKey(result.iteration, bytearray(result.seed))

        return None

    def getStructure(self):
        return self.senderKeyStateStructure



from .ciphertextmessage import CiphertextMessage
from ..util.byteutil import ByteUtil
from ..ecc.curve import Curve
from ..identitykey import IdentityKey
from .whispermessage import WhisperMessage
from axolotl.invalidversionexception import InvalidVersionException
from axolotl.invalidmessageexception import InvalidMessageException
from axolotl.legacymessageexception import LegacyMessageException
from axolotl.invalidkeyexception import InvalidKeyException
from google.protobuf.message import DecodeError
from . import whisperprotos
class PreKeyWhisperMessage(CiphertextMessage):
    def __init__(self, messageVersion = None, registrationId = None, preKeyId = None,
                 signedPreKeyId = None, ecPublicBaseKey = None, identityKey = None, whisperMessage = None, serialized = None):

        if serialized:
            try:
                self.version = ByteUtil.highBitsToInt(serialized[0])
                if self.version > CiphertextMessage.CURRENT_VERSION:
                    raise InvalidVersionException("Unknown version %s" % self.version)

                preKeyWhisperMessage = whisperprotos.PreKeyWhisperMessage()
                preKeyWhisperMessage.ParseFromString(serialized[1:])

                if (self.version == 2 and preKeyWhisperMessage.preKeyId is None) or\
                    (self.version == 3 and preKeyWhisperMessage.signedPreKeyId is None) or\
                    not preKeyWhisperMessage.baseKey or\
                    not preKeyWhisperMessage.identityKey or \
                    not preKeyWhisperMessage.message:

                    raise InvalidMessageException("Incomplete message")

                self.serialized = serialized
                self.registrationId = preKeyWhisperMessage.registrationId
                self.preKeyId       = preKeyWhisperMessage.preKeyId
                self.signedPreKeyId = preKeyWhisperMessage.signedPreKeyId if preKeyWhisperMessage.signedPreKeyId is not None else -1
                self.baseKey        = Curve.decodePoint(bytearray(preKeyWhisperMessage.baseKey), 0)

                self.identityKey    = IdentityKey(Curve.decodePoint(bytearray(preKeyWhisperMessage.identityKey), 0))
                self.message =  WhisperMessage(serialized = preKeyWhisperMessage.message)
            except (InvalidKeyException, LegacyMessageException, DecodeError) as e:
                raise InvalidMessageException(e)

        else:
            self.version = messageVersion
            self.registrationId = registrationId
            self.preKeyId = preKeyId
            self.signedPreKeyId = signedPreKeyId
            self.baseKey = ecPublicBaseKey
            self.identityKey = identityKey
            self.message = whisperMessage

            builder = whisperprotos.PreKeyWhisperMessage()
            builder.signedPreKeyId = signedPreKeyId
            builder.baseKey = ecPublicBaseKey.serialize()
            builder.identityKey = identityKey.serialize()
            builder.message = whisperMessage.serialize()
            builder.registrationId = registrationId

            if preKeyId is not None:
                builder.preKeyId = preKeyId

            versionBytes = ByteUtil.intsToByteHighAndLow(self.version, self.__class__.CURRENT_VERSION)
            messageBytes = builder.SerializeToString()
            self.serialized = bytes(ByteUtil.combine(versionBytes, messageBytes))

    def getMessageVersion(self):
        return self.version

    def getIdentityKey(self):
        return self.identityKey

    def getRegistrationId(self):
        return self.registrationId

    def getPreKeyId(self):
        return self.preKeyId

    def getSignedPreKeyId(self):
        return self.signedPreKeyId

    def getBaseKey(self):
        return self.baseKey

    def getWhisperMessage(self):
        return self.message

    def serialize(self):
        return self.serialized

    def getType(self):
        return CiphertextMessage.PREKEY_TYPE



from axolotl.state.storageprotos import SenderKeyRecordStructure
from axolotl.groups.state.senderkeystate import SenderKeyState
from axolotl.invalidkeyidexception import InvalidKeyIdException

class SenderKeyRecord:
    def __init__(self, serialized = None):
        self.senderKeyStates = []

        if serialized:
            senderKeyRecordStructure = SenderKeyRecordStructure()
            senderKeyRecordStructure.ParseFromString(serialized)

            for structure in senderKeyRecordStructure.senderKeyStates:
                self.senderKeyStates.append(SenderKeyState(senderKeyStateStructure=structure))


    def getSenderKeyState(self, keyId = None):
        if keyId is None:
            if len(self.senderKeyStates):
                return self.senderKeyStates[0]
            else:
                raise InvalidKeyIdException("No key state in record")
        else:
            for state in self.senderKeyStates:
                if state.getKeyId() == keyId:
                    return state
            raise InvalidKeyIdException("No keys for: %s" % keyId)


    def addSenderKeyState(self, id, iteration, chainKey, signatureKey):
        """
        :type id: int
        :type iteration: int
        :type chainKey: bytearray
        :type signatureKey: ECPublicKey
        """
        self.senderKeyStates.append(SenderKeyState(id, iteration, chainKey, signatureKey))

    def setSenderKeyState(self, id, iteration, chainKey, signatureKey):
        """
        :type id: int
        :type iteration: int
        :type chainKey: bytearray
        :type signatureKey: ECKeyPair
        """
        del self.senderKeyStates[:]
        self.senderKeyStates.append(SenderKeyState(id, iteration, chainKey, signatureKeyPair=signatureKey))

    def serialize(self):
        recordStructure = SenderKeyRecordStructure()

        for senderKeyState in self.senderKeyStates:
            recordStructure.senderKeyStates.extend([senderKeyState.getStructure()])

        return recordStructure.SerializeToString()
